"""
    secret_lock_transaction
    ========================

    Secret-lock transaction, which locks a given mosaic.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .hash_type import HashType
from .inner_transaction import InnerTransaction
from .recipient import Recipient, RecipientType
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ..mosaic.mosaic import Mosaic
from ..mosaic.mosaic_id import MosaicId
from ... import util

__all__ = [
    'SecretLockTransaction',
    'SecretLockInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('SECRET_LOCK')
class SecretLockTransaction(Transaction):
    """
    Secret-proof transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param mosaic: Mosaic to be locked.
    :param duration: Duration for funds to be released or returned.
    :param hash_type: Hash algorithm secret was generated with.
    :param secret: Hex-encoded or raw seed-proof hash.
    :param recipient: Recipient of funds.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    mosaic: Mosaic
    duration: int
    hash_type: HashType
    secret: str
    recipient: RecipientType

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        mosaic: Mosaic,
        duration: int,
        hash_type: HashType,
        secret: str,
        recipient: RecipientType,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        secret = util.encode_hex(secret)
        if not hash_type.validate(secret):
            raise ValueError("HashType and secret have incompatible lengths.")
        super().__init__(
            TransactionType.SECRET_LOCK,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('mosaic', mosaic)
        self._set('duration', duration)
        self._set('hash_type', hash_type)
        self._set('secret', secret)
        self._set('recipient', recipient)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        mosaic: Mosaic,
        duration: int,
        hash_type: HashType,
        secret: str,
        recipient: RecipientType,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new secret proof transaction.

        :param deadline: Deadline to include transaction.
        :param mosaic: Mosaic to be locked.
        :param duration: Duration for funds to be released or returned.
        :param hash_type: Hash algorithm secret was generated with.
        :param secret: Hex-encoded or raw seed-proof hash.
        :param recipient: Recipient of funds.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.SECRET_LOCK,
            deadline,
            mosaic,
            duration,
            hash_type,
            secret,
            recipient,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        mosaic_size = Mosaic.CATBUFFER_SIZE
        duration_size = util.U64_BYTES
        hash_size = util.U8_BYTES
        secret_size = 32
        recipient_size = 25

        return mosaic_size + duration_size + hash_size + secret_size + recipient_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export secret lock-specific data to catbuffer."""

        # Mosaic mosaic
        # uint64_t duration
        # uint8_t hash_type
        # uint8_t[32] secret
        # uint8_t[25] recipient
        mosaic = self.mosaic.to_catbuffer(network_type)
        duration = util.u64_to_catbuffer(self.duration)
        hash_type = self.hash_type.to_catbuffer(network_type)
        secret = util.unhexlify(self.secret)
        secret = secret + b'\x00' * (32 - len(secret))
        recipient = Recipient.to_catbuffer(self.recipient, network_type)
        return mosaic + duration + hash_type + secret + recipient

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load secret proof-specific data data from catbuffer."""

        # Mosaic mosaic
        # uint64_t duration
        # uint8_t hash_type
        # uint8_t[32] secret
        # uint8_t[25] recipient
        mosaic, data = Mosaic.create_from_catbuffer_pair(data, network_type)
        duration = util.u64_from_catbuffer(data[:8])
        hash_type, data = HashType.create_from_catbuffer_pair(data[8:], network_type)
        hash_length = hash_type.hash_length() // 2
        secret = util.hexlify(data[:hash_length])
        data = data[32:]
        recipient, data = Recipient.create_from_catbuffer_pair(data, network_type)

        self._set('mosaic', mosaic)
        self._set('duration', duration)
        self._set('hash_type', hash_type)
        self._set('secret', secret)
        self._set('recipient', recipient)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {
            'mosaicId',
            'amount',
            'duration',
            'hashAlgorithm',
            'secret',
            'recipient',
        }
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'mosaicId': util.u64_to_dto(int(self.mosaic.id)),
            'amount': util.u64_to_dto(self.mosaic.amount),
            'duration': util.u64_to_dto(self.duration),
            'hashAlgorithm': self.hash_type.to_dto(network_type),
            'secret': self.secret,
            'recipient': Recipient.to_dto(self.recipient, network_type),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        mosaic_id = MosaicId(util.u64_from_dto(data['mosaicId']))
        amount = util.u64_from_dto(data['amount'])
        mosaic = Mosaic(mosaic_id, amount)
        duration = util.u64_from_dto(data['duration'])
        hash_type = HashType.create_from_dto(data['hashAlgorithm'], network_type)
        recipient = Recipient.create_from_dto(data['recipient'], network_type)

        self._set('mosaic', mosaic)
        self._set('duration', duration)
        self._set('hash_type', hash_type)
        self._set('secret', data['secret'])
        self._set('recipient', recipient)


@register_transaction('SECRET_LOCK')
class SecretLockInnerTransaction(InnerTransaction, SecretLockTransaction):
    """Embedded secret proof transaction."""

    __slots__ = ()



from xpxchain.util.signature import ed25519
from tests import harness


class TestEd25519(harness.TestCase):

    @harness.randomize(
        calls=100,
        seed={'fixed_length': 32},
        message={'min_length': 0, 'max_length': 100}
    )
    def test_sha2(self, seed: bytes, message: bytes):
        sk1 = ed25519.sha2.SigningKey(seed)
        vk1 = sk1.get_verifying_key()
        sig1 = sk1.sign(message)

        # Test success.
        self.assertEqual(sk1.to_seed(), seed)
        vk1.verify(sig1, message)

        # Test invalid key.
        sk2 = ed25519.sha2.SigningKey(seed[16:] + seed[:16])
        vk2 = sk2.get_verifying_key()
        self.assertNotEqual(sk2.to_seed(), seed)
        with self.assertRaises(ed25519.sha2.BadSignatureError):
            vk2.verify(sig1, message)

        # Test invalid signature.
        sig2 = sig1[32:] + sig1[:32]
        with self.assertRaises(ed25519.sha2.BadSignatureError):
            vk1.verify(sig2, message)

    @harness.randomize(
        calls=100,
        seed={'fixed_length': 32},
        message={'min_length': 0, 'max_length': 100}
    )
    def test_sha3(self, seed: bytes, message: bytes):
        sk1 = ed25519.sha3.SigningKey(seed)
        vk1 = sk1.get_verifying_key()
        sig1 = sk1.sign(message)

        # Test success.
        self.assertEqual(sk1.to_seed(), seed)
        vk1.verify(sig1, message)

        # Test invalid key.
        sk2 = ed25519.sha3.SigningKey(seed[16:] + seed[:16])
        vk2 = sk2.get_verifying_key()
        self.assertNotEqual(sk2.to_seed(), seed)
        with self.assertRaises(ed25519.sha3.BadSignatureError):
            vk2.verify(sig1, message)

        # Test invalid signature.
        sig2 = sig1[32:] + sig1[:32]
        with self.assertRaises(ed25519.sha3.BadSignatureError):
            vk1.verify(sig2, message)



from xpxchain import util
from tests import harness


class TestIdentifier(harness.TestCase):

    def test_generate_mosaic_id(self):
        nonce = b'\x00\x00\x00\x00'
        public_key = util.unhexlify('7D08373CFFE4154E129E04F0827E5F3D6907587E348757B0F87D2F839BF88246')
        self.assertEqual(util.generate_mosaic_id(nonce, public_key), 0x2FF7D64F483BC0A6)

    def test_generate_sub_namespace_id(self):
        self.assertEqual(util.generate_sub_namespace_id(0, "sample"), 0x88B64C3BE2F47144)
        self.assertEqual(util.generate_sub_namespace_id(0x88B64C3BE2F47144, "sub"), 0xFA9429715A71ACC9)
        self.assertEqual(util.generate_sub_namespace_id(0xFA9429715A71ACC9, "full"), 0x8BC7011B0B344C54)

    def test_generate_namespace_id(self):
        ids = [0x88B64C3BE2F47144, 0xFA9429715A71ACC9, 0x8BC7011B0B344C54]
        self.assertEqual(util.generate_namespace_id(""), ids[:0])
        self.assertEqual(util.generate_namespace_id("sample"), ids[:1])
        self.assertEqual(util.generate_namespace_id("sample.sub"), ids[:2])
        self.assertEqual(util.generate_namespace_id("sample.sub.full"), ids)

        with self.assertRaises(ValueError):
            util.generate_namespace_id("sample.sub.full.overload")
        with self.assertRaises(ValueError):
            util.generate_namespace_id("-sample")
        with self.assertRaises(ValueError):
            util.generate_namespace_id("_sample")
        with self.assertRaises(ValueError):
            util.generate_namespace_id("sample*")



"""
    multisig_account_info
    =====================

    Describes multisig account and the levels the account is involved in.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from .address import Address
from .public_account import PublicAccount, PublicAccountList
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['MultisigAccountInfo']


@util.inherit_doc
@util.dataclass(frozen=True)
class MultisigAccountInfo(util.DTO):
    """
    Information describing a multisig account.

    Multisig accounts allow multiple accounts, or signatories,
    operate as an entity, requiring a certain consensus for operations.

    :param account: Multisig public account.
    :param min_approval: Min number of cosignatories required to approve a transaction.
    :param min_removal: Min number of cosignatories required to remove a cosignatory.
    :param cosignatories: List of cosignatories.
    :param multisig_accounts: List of multisig accounts this account cosigns.


    DTO Format:
        .. code-block:: yaml

            MultisigDTO:
                # Hex(PublicKey) (64-bytes)
                account: string
                # Hex(Address) (50-bytes)
                accountAddress: string
                minApproval: integer
                minRemoval: integer
                # Hex(PublicKey)[] (64-bytes)
                cosignatories: string[]
                # Hex(PublicKey)[] (64-bytes)
                multisigAccounts: string[]

            MultisigAccountInfoDTO:
                multisig: MultisigDTO
    """

    account: PublicAccount
    min_approval: int
    min_removal: int
    cosignatories: PublicAccountList
    multisig_accounts: PublicAccountList

    def is_multisig(self) -> bool:
        """Get if account is a multisig account."""

        return self.min_removal != 0 and self.min_approval != 0

    def has_cosigner(self, account: PublicAccount) -> bool:
        """
        Check if another account is cosignatory of multisig account.

        :param account: Other account.
        """

        return account in self.cosignatories

    def is_cosigner_of_multisig_account(self, account: PublicAccount) -> bool:
        """
        Check if multisig account is cosignatory of another account.

        :param account: Other account.
        """

        return account in self.multisig_accounts

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'multisig'}
        required_l2 = {
            'account',
            'accountAddress',
            'minApproval',
            'minRemoval',
            'cosignatories',
            'multisigAccounts',
        }
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
            # Level 2
            and cls.validate_dto_required(data['multisig'], required_l2)
            and cls.validate_dto_all(data['multisig'], required_l2)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        cosignatories = [i.public_key for i in self.cosignatories]
        multisig_accounts = [i.public_key for i in self.multisig_accounts]
        return {
            'multisig': {
                'account': self.account.public_key,
                'accountAddress': util.hexlify(self.account.address.encoded),
                'minApproval': util.u32_to_dto(self.min_approval),
                'minRemoval': util.u32_to_dto(self.min_removal),
                'cosignatories': cosignatories,
                'multisigAccounts': multisig_accounts,
            }
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        # Normalize the network type if it's not provided.
        if network_type is None and 'accountAddress' in data:
            network_type = Address(data['accountAddress']).network_type
        assert network_type is not None

        multisig = data['multisig']
        create_account = lambda x: PublicAccount.create_from_public_key(x, network_type)
        cosignatories = [create_account(i) for i in multisig['cosignatories']]
        multisig_accounts = [create_account(i) for i in multisig['multisigAccounts']]
        return cls(
            account=create_account(multisig['account']),
            min_approval=util.u32_from_dto(multisig['minApproval']),
            min_removal=util.u32_from_dto(multisig['minRemoval']),
            cosignatories=cosignatories,
            multisig_accounts=multisig_accounts
        )



"""
    binascii
    ========

    Utilities for converting between binary and ascii data.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import binascii
import typing


__all__ = [
    'decode_hex',
    'encode_hex',
    'hexlify',
    'unhexlify',
]


def hexlify(data: bytes, with_prefix=False) -> str:
    """Convert bytes data to hexadecimal string."""

    encoded = binascii.hexlify(data).decode('ascii')
    if with_prefix:
        return '0x' + encoded
    return encoded


def encode_hex(data: typing.AnyStr, with_prefix=False) -> str:
    """Encodes raw bytes to hex."""

    if isinstance(data, str):
        return data
    return hexlify(data, with_prefix=with_prefix)


def unhexlify(data: typing.AnyStr, with_prefix=False) -> bytes:
    """Convert hexadecimal string to raw bytes."""

    if with_prefix:
        data = remove_prefix(data)
    return binascii.unhexlify(data)


def decode_hex(data: typing.AnyStr, with_prefix=False) -> bytes:
    """Decode hex data to raw bytes."""

    if isinstance(data, (bytes, bytearray)):
        return data
    return unhexlify(data, with_prefix=with_prefix)


@typing.no_type_check
def remove_prefix(data: typing.AnyStr) -> typing.AnyStr:
    """Remove hexadecimal prefix from string ('0x', '0X')"""

    if isinstance(data, str):
        prefix = ('0x', '0X')
    else:
        prefix = (b'0x', b'0X')

    if data.startswith(prefix):
        return data[2:]
    return data



"""
    network_type
    ============

    Constants for the network type.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import enum
import typing

from ... import util


__all__ = ['NetworkType']


@util.inherit_doc
class NetworkType(util.U8Mixin, util.EnumMixin, enum.IntEnum):
    """
    Identifier for the network type.

    DTO Format:
        .. code-block:: yaml

            NetworkTypeDTO:
                name: string
                description: string
    """

    MAIN_NET   = 0xb8       # noqa: E221
    TEST_NET   = 0xa8       # noqa: E221
    MIJIN      = 0x60       # noqa: E221
    MIJIN_TEST = 0x90       # noqa: E221

    def description(self) -> str:
        return DESCRIPTION[self]

    def identifier(self) -> bytes:
        """Get address identifier from type."""
        return TO_IDENTIFIER[self]

    @classmethod
    def create_from_identifier(cls, identifier: bytes):
        """
        Identify and create the network type from the raw address identifier.

        :param identifier: First character of the raw address.
        """

        assert len(identifier) == 1
        return FROM_IDENTIFIER[identifier]

    @classmethod
    def create_from_raw_address(cls, address: str):
        """
        Identify and create the network type from the raw address.

        :param address: Base32-decoded, upper-case, stripped address.
        """

        assert len(address) == 40
        return cls.create_from_identifier(address[0].encode('ascii'))


DESCRIPTION = {
    NetworkType.MAIN_NET: "Main network",
    NetworkType.TEST_NET: "Test network",
    NetworkType.MIJIN: "Mijin network",
    NetworkType.MIJIN_TEST: "Mijin test network",
}

TO_IDENTIFIER = {
    NetworkType.MAIN_NET: b"X",
    NetworkType.TEST_NET: b"V",
    NetworkType.MIJIN: b"M",
    NetworkType.MIJIN_TEST: b"S",
}

FROM_IDENTIFIER = {
    b"X": NetworkType.MAIN_NET,
    b"V": NetworkType.TEST_NET,
    b"M": NetworkType.MIJIN,
    b"S": NetworkType.MIJIN_TEST,
}

OptionalNetworkType = typing.Optional[NetworkType]



"""
    namespace_name
    ==============

    Describes a namespace by name and identifier.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import re
import typing

from .namespace_id import NamespaceId
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['NamespaceName']


@util.inherit_doc
@util.dataclass(frozen=True, parent_id=None)
class NamespaceName(util.DTO):
    """
    Namespace name and identifier.

    :param namespace_id: Namespace ID.
    :param name: Namespace name.

    DTO Format:
        .. code-block:: yaml

            NamespaceNameDTO:
                namespaceId: UInt64DTO
                name: string
                parentId?: UInt64DTO
    """

    namespace_id: NamespaceId
    name: str
    parent_id: typing.Optional[NamespaceId]

    @classmethod
    def create_from_name(cls, name: str):
        """
        Create namespace name and identifier from name.

        :param name: Namespace name.
        """

        ids = util.generate_namespace_id(name)
        if len(ids) == 0:
            namespace_id = NamespaceId(0)
            parent_id = None
        elif len(ids) == 1:
            namespace_id = NamespaceId(ids[0])
            parent_id = None
        else:
            namespace_id = NamespaceId(ids[-1])
            parent_id = NamespaceId(ids[-2])

        return cls(
            namespace_id=namespace_id,
            name=name.split('.')[-1],
            parent_id=parent_id,
        )

    def is_valid(self) -> bool:
        """Determine if the namespace name is valid."""
        return (
            len(self.name) <= 64
            and re.match(r'\A[a-z0-9_-]+\Z', self.name) is not None
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        data = {
            'namespaceId': util.u64_to_dto(int(self.namespace_id)),
            'name': self.name
        }
        if self.parent_id is not None:
            data['parentId'] = util.u64_to_dto(int(self.parent_id))

        return data

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        parent_id = None
        if 'parentId' in data:
            parent_id = NamespaceId(util.u64_from_dto(data['parentId']))
        return cls(
            namespace_id=NamespaceId(util.u64_from_dto(data['namespaceId'])),
            name=data['name'],
            parent_id=parent_id,
        )



"""
    mosaic_supply_change_transaction
    ================================

    Mosaic supply change transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ..mosaic.mosaic_id import MosaicId
from ..mosaic.mosaic_supply_type import MosaicSupplyType
from ... import util

__all__ = [
    'MosaicSupplyChangeTransaction',
    'MosaicSupplyChangeInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('MOSAIC_SUPPLY_CHANGE')
class MosaicSupplyChangeTransaction(Transaction):
    """
    Mosaic supply change transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.

    :param nonce: Mosaic nonce (random data for mosaic uniqueness).
    :param mosaic_id: Identifier for mosaic.
    :param mosaic_properties: Mosaic properties.

    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    mosaic_id: MosaicId
    direction: MosaicSupplyType
    delta: int

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        mosaic_id: MosaicId,
        direction: MosaicSupplyType,
        delta: int,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        super().__init__(
            TransactionType.MOSAIC_SUPPLY_CHANGE,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('mosaic_id', mosaic_id)
        self._set('direction', direction)
        self._set('delta', delta)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        mosaic_id: MosaicId,
        direction: MosaicSupplyType,
        delta: int,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new mosaic supply change transaction.

        :param deadline: Deadline to include transaction.

        :param nonce: Mosaic nonce (random data for mosaic uniqueness).
        :param mosaic_id: Identifier for mosaic.
        :param mosaic_properties: Mosaic properties.

        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.MOSAIC_SUPPLY_CHANGE,
            deadline,
            mosaic_id,
            direction,
            delta,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        id_size = util.U64_BYTES
        direction_size = MosaicSupplyType.CATBUFFER_SIZE
        delta_size = util.U64_BYTES
        return id_size + direction_size + delta_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export mosaic supply change-specific data to catbuffer."""

        # uint64 mosaic_id
        # MosaicSupplyChange direction
        # uint64 delta
        mosaic_id = util.u64_to_catbuffer(int(self.mosaic_id))
        direction = self.direction.to_catbuffer(network_type)
        delta = util.u64_to_catbuffer(self.delta)
        return mosaic_id + direction + delta

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load mosaic supply change-specific data data from catbuffer."""

        # uint64 mosaic_id
        # MosaicSupplyChange direction
        # uint64 delta
        mosaic_id = MosaicId(util.u64_from_catbuffer(data[:8]))
        direction = MosaicSupplyType.create_from_catbuffer(data[8:9], network_type)
        delta = util.u64_from_catbuffer(data[9:17])
        data = data[17:]

        self._set('mosaic_id', mosaic_id)
        self._set('direction', direction)
        self._set('delta', delta)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'mosaicId', 'direction', 'delta'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'mosaicId': util.u64_to_dto(int(self.mosaic_id)),
            'direction': self.direction.to_dto(network_type),
            'delta': util.u64_to_dto(self.delta),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        mosaic_id = MosaicId(util.u64_from_dto(data['mosaicId']))
        direction = MosaicSupplyType.create_from_dto(data['direction'], network_type)
        delta = util.u64_from_dto(data['delta'])
        self._set('mosaic_id', mosaic_id)
        self._set('direction', direction)
        self._set('delta', delta)


@register_transaction('MOSAIC_SUPPLY_CHANGE')
class MosaicSupplyChangeInnerTransaction(InnerTransaction, MosaicSupplyChangeTransaction):
    """Embedded mosaic supply change transaction."""

    __slots__ = ()



"""
    receipt_type
    ================

    Enumerations for transaction types.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import enum

from ... import util

__all__ = ['ReceiptType']


@util.inherit_doc
class ReceiptType(util.U16Mixin, util.EnumMixin, enum.IntEnum):
    """Receipt type."""

    MOSAIC_RENTAL_FEE = 0x134D
    NAMESPACE_RENTAL_FEE = 0x124E
    VALIDATE_FEE = 0x2143
    MOSAIC_LEVY = 0x124D
    LOCKHASH_COMPLETED = 0x2248
    LOCKHASH_EXPIRED = 0x2348
    LOCKSECRET_COMPLETED = 0x2252
    LOCKSECRET_EXPIRED = 0x2352
    LOCKHASH_CREATED = 0x3148
    LOCKSECRET_CREATED = 0x3152
    MOSAIC_EXPIRED = 0x414D
    NAMESPACE_EXPIRED = 0x414E
    INFLATION = 0x5143
    TRANSACTION_GROUP = 0xE134
    ADDRESS_ALIAS_RESOLUTION = 0xF143
    MOSAIC_ALIAS_RESOLUTION = 0xF243

    def description(self) -> str:
        return DESCRIPTION[self]


DESCRIPTION = {

    ReceiptType.MOSAIC_RENTAL_FEE: "Mosaic rental fee receipt type.",
    ReceiptType.NAMESPACE_RENTAL_FEE: "Namespace rental fee receipt type.",
    ReceiptType.VALIDATE_FEE: "Harvest fee receipt type.",
    ReceiptType.LOCKHASH_COMPLETED: "Lock hash completed receipt type.",
    ReceiptType.LOCKHASH_EXPIRED: "Lock hash expired receipt type.",
    ReceiptType.LOCKSECRET_COMPLETED: "Lock secret completed receipt type.",
    ReceiptType.LOCKSECRET_EXPIRED: "Lock secret expired receipt type.",
    ReceiptType.LOCKHASH_CREATED: "Lock hash created receipt type.",
    ReceiptType.LOCKSECRET_CREATED: "Lock secret created receipt type.",
    ReceiptType.MOSAIC_EXPIRED: "Mosaic expired receipt type.",
    ReceiptType.NAMESPACE_EXPIRED: "Namespace expired receipt type.",
    ReceiptType.INFLATION: "Inflation receipt type.",
    ReceiptType.TRANSACTION_GROUP: "Transaction group receipt type.",
    ReceiptType.ADDRESS_ALIAS_RESOLUTION: "Address alias resolution receipt type.",
    ReceiptType.MOSAIC_ALIAS_RESOLUTION: "Mosaic alias resolution receipt type.",

}



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses


@harness.mocked_http_test_case({
    'clients': (client.MosaicHTTP, client.AsyncMosaicHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_mosaic',
            'response': responses.MOSAIC_INFO["Ok"],
            'params': [models.MosaicId.create_from_hex('6c699a1517bea955')],
            'method': 'get_mosaic',
            'validation': [
                lambda x: (x.meta_id, '5CC07CBC3A48065F47D6DF80'),
                lambda x: (hex(x.mosaic_id), '0x6c699a1517bea955'),
                lambda x: (x.supply, 8999999998000000),
                lambda x: (x.height, 1),
                lambda x: (x.owner.public_key, "A04335F99D9EE3787528A16C7A302F80D511E9CF71D97D95C2182E0EA75A1EF9"),
                lambda x: (x.revision, 1),
                lambda x: (x.properties.flags, 0x2),
                lambda x: (x.properties.divisibility, 6),
                lambda x: (x.properties.duration, 0),
            ]
        },
        {
            'name': 'test_get_mosaics',
            'response': responses.MOSAICS_INFO["Ok"],
            'params': [[models.MosaicId.create_from_hex('6c699a1517bea955')]],
            'method': 'get_mosaics',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].meta_id, '5CC07CBC3A48065F47D6DF80'),
            ]
        },
        {
            'name': 'test_get_mosaic_names',
            'response': responses.MOSAICS_NAMES["Ok"],
            'params': [[models.MosaicId.create_from_hex('0dc67fbe1cad29e3')]],
            'method': 'get_mosaic_names',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (isinstance(x[0], models.MosaicName), True),
                lambda x: (isinstance(x[0].mosaic_id, models.MosaicId), True),
                lambda x: (int(x[0].mosaic_id), 0x0dc67fbe1cad29e3),
                lambda x: (len(x[0].names), 1),
                lambda x: (x[0].names[0], 'prx.xpx'),
            ]
        },
    ],
})
class TestMosaicHTTP(harness.TestCase):
    pass



import binascii

from xpxchain import util
from tests import harness


class HexTest(harness.TestCase):

    def test_hexlify(self):
        self.assertEqual(util.hexlify(b'5'), '35')
        self.assertEqual(util.hexlify(b'5', with_prefix=True), '0x35')

        with self.assertRaises(TypeError):
            util.hexlify('5')

    def test_unhexlify(self):
        self.assertEqual(util.unhexlify('35'), b'5')
        self.assertEqual(util.unhexlify(b'35'), b'5')
        self.assertEqual(util.unhexlify('0x35', with_prefix=True), b'5')
        self.assertEqual(util.unhexlify(b'0x35', with_prefix=True), b'5')

        with self.assertRaises(binascii.Error):
            util.unhexlify('5')
        with self.assertRaises(binascii.Error):
            util.unhexlify('0x35')



"""
    mosaic_info
    ===========

    Detailed information describing an asset.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from .mosaic_id import MosaicId
from .mosaic_properties import MosaicProperties
from ..account.public_account import PublicAccount
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['MosaicInfo']


@util.inherit_doc
@util.dataclass(frozen=True)
class MosaicInfo(util.DTO):
    """
    Information describing a mosaic.

    :param active: Mosaic is active.
    :param index: Mosaic index.
    :param meta_id: Mosaic metadata ID.
    :param mosaic_id: Mosaic ID.
    :param nonce: Mosaic nonce.
    :param supply: Mosaic supply.
    :param height: Block height when mosaic was created.
    :param owner: Account that owns mosaic.
    :param properties: Mosaic properties.

    DTO Format:
        .. code-block:: yaml

            MosaicMetaDTO:
                # Hex(Id) (24-bytes)
                id: string

            MosaicDefinitionDTO:
                mosaicId: UInt64DTO
                supply: UInt64DTO
                height: UInt64DTO
                # Hex(PublicKey) (64-bytes)
                owner: string
                revision: integer
                properties: MosaicPropertiesDTO

            MosaicInfoDTO:
                meta: MosaicMetaDTO
                mosaic: MosaicDefinitionDTO
    """

    meta_id: str
    mosaic_id: MosaicId
    supply: int
    height: int
    owner: PublicAccount
    revision: int
    properties: MosaicProperties

    @property
    def divisibility(self) -> int:
        """Get the decimal place mosaic can be divided into."""
        return self.properties.divisibility

    @property
    def duration(self) -> int:
        """Get the number of blocks the mosaic will be available."""
        return self.properties.duration

    def is_supply_mutable(self) -> bool:
        """Mosaic allows a supply change later on. Default false."""
        return self.properties.supply_mutable

    def is_transferable(self) -> bool:
        """Allow transfer of funds from non-creator accounts. Default true."""
        return self.properties.transferable

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'meta', 'mosaic'}
        required_l21 = {'id'}
        required_l22 = {
            'mosaicId',
            'supply',
            'height',
            'owner',
            'revision',
            'properties',
        }
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
            # Level 2_1
            and cls.validate_dto_required(data['meta'], required_l21)
            and cls.validate_dto_all(data['meta'], required_l21)
            # Level 2_2
            and cls.validate_dto_required(data['mosaic'], required_l22)
            # and cls.validate_dto_all(data['mosaic'], required_l22)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'meta': {
                'id': self.meta_id,
            },
            'mosaic': {
                'mosaicId': util.u64_to_dto(int(self.mosaic_id)),
                'supply': util.u64_to_dto(self.supply),
                'height': util.u64_to_dto(self.height),
                'owner': self.owner.public_key,
                'revision': util.u32_to_dto(self.revision),
                'properties': self.properties.to_dto(network_type),
            },
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        meta_dto = data['meta']
        mosaic_dto = data['mosaic']
        owner_dto = mosaic_dto['owner']
        return cls(
            meta_id=meta_dto['id'],
            mosaic_id=MosaicId(util.u64_from_dto(mosaic_dto['mosaicId'])),
            supply=util.u64_from_dto(mosaic_dto['supply']),
            height=util.u64_from_dto(mosaic_dto['height']),
            owner=PublicAccount.create_from_public_key(owner_dto, network_type),
            revision=util.u32_from_dto(mosaic_dto['revision']),
            properties=MosaicProperties.create_from_dto(mosaic_dto['properties']),
        )



"""
    alias_transaction
    =================

    Abstract class for alias transactions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from .address_alias_transaction import AddressAliasTransaction
from .deadline import Deadline
from .mosaic_alias_transaction import MosaicAliasTransaction
from .transaction import TransactionBase
from ..account.address import Address
from ..blockchain.network_type import NetworkType
from ..mosaic.mosaic_id import MosaicId
from ..namespace.alias_action_type import AliasActionType
from ..namespace.namespace_id import NamespaceId

__all__ = ['AliasTransaction']


class AliasTransaction(TransactionBase):
    """Abstract class for alias transactions."""

    __slots__ = ()

    @staticmethod
    def create_for_address(
        deadline: Deadline,
        action_type: AliasActionType,
        namespace_id: NamespaceId,
        address: Address,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new address alias transaction.

        :param deadline: Deadline to include transaction.
        :param action_type: Alias action type.
        :param namespace_id: Resulting namespace ID that will be an alias.
        :param address: Address to be aliased.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """

        return AddressAliasTransaction.create(
            deadline,
            action_type,
            namespace_id,
            address,
            network_type,
            max_fee,
        )

    @staticmethod
    def create_for_mosaic(
        deadline: Deadline,
        action_type: AliasActionType,
        namespace_id: NamespaceId,
        mosaic_id: MosaicId,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new mosaic alias transaction.

        :param deadline: Deadline to include transaction.
        :param action_type: Alias action type.
        :param namespace_id: Resulting namespace ID that will be an alias.
        :param mosaic_id: Mosaic to be aliased.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """

        return MosaicAliasTransaction.create(
            deadline,
            action_type,
            namespace_id,
            mosaic_id,
            network_type,
            max_fee,
        )



from xpxchain import models
from xpxchain import util
from tests import harness


@harness.model_test_case({
    'type': models.Account,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'address': models.Address('SAUJCIBCOFLHUZIWNB32MR6YUX75HO7GGCVZEXSG'),
        'public_key': '1b153f8b76ef60a4bfe152f4de3698bd230bac9dc239d4e448715aa46bd58955',
        'private_key': '97131746d864f4c9001b1b86044d765ba08d7fddc7a0fb3abbc8d111aa26cdca',
    },
})
class TestAccount(harness.TestCase):

    def test_properties(self):
        public_account = models.PublicAccount(self.model.address, self.model.public_key)
        self.assertEqual(self.model.network_type, self.network_type)
        self.assertEqual(self.model.public_account, public_account)

    @harness.ignore_warnings_test
    def test_create_from_private_key(self):
        value = self.type.create_from_private_key(self.data['private_key'], self.network_type)
        self.assertEqual(value, self.model)

    @harness.ignore_warnings_test
    def test_generate_new_account(self):
        def fake_entropy(size: int) -> bytes:
            return b'0' * size

        # try the random version
        value = self.type.generate_new_account(self.network_type)
        message = b'Hello World!'
        signature = value.sign_data(message)
        self.assertTrue(value.public_account.verify_signature(message, signature))

        # try with a deterministic version
        value = self.type.generate_new_account(self.network_type, fake_entropy)
        self.assertEqual(value.address.address, 'SA4PGCTZHFCTC5WHYSZCEJZPOO3X7OFBAHCMSL4U')
        self.assertEqual(value.public_key, 'dcedb3902b8fa33f8ced2c9ea62497ae4f43e8208cdedaf4d7cac50abc53ddac')
        self.assertEqual(value.private_key, '3030303030303030303030303030303030303030303030303030303030303030')
        signature = util.hexlify(value.sign_data(message))
        self.assertEqual(signature, 'bbe720fa06fe26f79896f5880baf7e8d112d796c648c56414e2079349baff54c5d46ab08115fad1865de2288a1eeecfaedc3143df14fc81841aec17fa8cba70e')
        self.assertTrue(value.public_account.verify_signature(message, signature))

    @harness.ignore_warnings_test
    def test_sign(self):
        transaction = 'bb000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000019052420000000000000000f1b4815c00000000009b3155b37159da50aa52d5967c509b410f5a36a3b1e31ecb5ac76675d79b4a5e2000b778a39a3663719dfc5e48c9d78431b1e45c2af9df538782bf199c189dabeac7'
        signed = util.hexlify(self.model.sign(transaction, '7B631D803F912B00DC0CBED3014BBD17A302BA50B99D233B9C2D9533B842ABDF'))
        self.assertEqual(signed, 'bb000000111f1e4ca0f97a1020e9339e9c21dddddd231a48c28d0b3eade6c9567d6092ee9a36490c2ea6482cd7cef7c4f07d49e5ba33d4b278bd4ff95e56e916f448890a1b153f8b76ef60a4bfe152f4de3698bd230bac9dc239d4e448715aa46bd58955019052420000000000000000f1b4815c00000000009b3155b37159da50aa52d5967c509b410f5a36a3b1e31ecb5ac76675d79b4a5e2000b778a39a3663719dfc5e48c9d78431b1e45c2af9df538782bf199c189dabeac7')

    @harness.ignore_warnings_test
    def test_sign_data(self):
        message = b'Hello World!'
        signature = util.hexlify(self.model.sign_data(message))
        self.assertEqual(signature, '40af0cb5a7a7533f07a4ba6f1cb2df64f2347feb1b2eaabb9374d28603d146497ea83d3d6ee15758d39c298b48f58e578cc42f36a373e15eef7412e0bd19a801')

    @harness.ignore_warnings_test
    def test_verify_signature(self):
        message = b'Hello World!'
        signature = '40af0cb5a7a7533f07a4ba6f1cb2df64f2347feb1b2eaabb9374d28603d146497ea83d3d6ee15758d39c298b48f58e578cc42f36a373e15eef7412e0bd19a801'
        self.assertTrue(self.model.verify_signature(message, signature))

    @harness.ignore_warnings_test
    def test_verify_transaction(self):
        payload = 'bb000000d0092d8eaf91c07069eeef6651cd313e792b27d2cb31473ceaac40f78ee2121acb5f665826083b87b374c9eb67aefef6b8cf74f0298820a9143b34055e15900c1b153f8b76ef60a4bfe152f4de3698bd230bac9dc239d4e448715aa46bd58955019052420000000000000000f1b4815c00000000009b3155b37159da50aa52d5967c509b410f5a36a3b1e31ecb5ac76675d79b4a5e2000b778a39a3663719dfc5e48c9d78431b1e45c2af9df538782bf199c189dabeac7'
        self.assertTrue(self.model.verify_transaction(payload))

    def test_invalid_keys(self):
        with self.assertRaises(ValueError):
            self.type(self.data['address'], self.data['public_key'], '')
        with self.assertRaises(ValueError):
            self.type(self.data['address'], '', self.data['private_key'])


@harness.model_test_case({
    'type': models.AccountInfo,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        # 'meta': models.AccountMeta(1, '40af0cb5a7a7533f07a4ba6f1cb2df64f2347feb1b2eaabb93', '74d28603d146497ea83d3d6ee15758d39c298b48f58', 1, 'e578'),
        'meta': models.AccountMeta(),
        'address': models.Address('SCBO3CAFOVAOGYBAHQKPUOGLAYWFLNJUFFCH3RYY'),
        'address_height': 1,
        'public_key': '7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808',
        'public_key_height': 1,
        'mosaics': [],
        'account_type': 0,
        'linked_account_key': '',
        'snapshots': [],
    },
    'dto': {
        'meta': {
            # 'height': [1, 0],
            # 'hash': '40af0cb5a7a7533f07a4ba6f1cb2df64f2347feb1b2eaabb93',
            # 'merkleComponentHash': '74d28603d146497ea83d3d6ee15758d39c298b48f58',
            # 'index': 1,
            # 'id': 'e578',
        },
        'account': {
            'address': '9082ed88057540e360203c14fa38cb062c55b53429447dc718',
            'addressHeight': [1, 0],
            'publicKey': '7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808',
            'publicKeyHeight': [1, 0],
            'mosaics': [],
            'accountType': 0,
            'linkedAccountKey': '',
            'snapshots': [],
        },
    }
})
class TestAccountInfo(harness.TestCase):

    def test_properties(self):
        public_account = models.PublicAccount(self.model.address, self.model.public_key)
        self.assertEqual(self.model.public_account, public_account)


@harness.model_test_case({
    'type': models.AccountMeta,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        # 'height': 1,
        # 'hash': '40af0cb5a7a7533f07a4ba6f1cb2df64f2347feb1b2eaabb93',
        # 'merkle_component_hash': '74d28603d146497ea83d3d6ee15758d39c298b48f58',
        # 'index': 1,
        # 'id': 'e578',
    },
    'dto': {
        # 'height': [1, 0],
        # 'hash': '40af0cb5a7a7533f07a4ba6f1cb2df64f2347feb1b2eaabb93',
        # 'merkleComponentHash': '74d28603d146497ea83d3d6ee15758d39c298b48f58',
        # 'index': 1,
        # 'id': 'e578',
    },
    'eq': False,
})
class TestAccountMeta(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.AccountProperty,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'property_type': models.PropertyType.ALLOW_ADDRESS,
        'values': [
            models.Address('SAUJCIBCOFLHUZIWNB32MR6YUX75HO7GGCVZEXSG'),
            models.Address('SCBO3CAFOVAOGYBAHQKPUOGLAYWFLNJUFFCH3RYY'),
        ],
    },
    'dto': {
        'propertyType': 0x01,
        'values': [
            '902891202271567A65166877A647D8A5FFD3BBE630AB925E46',
            '9082ED88057540E360203C14FA38CB062C55B53429447DC718',
        ],
    },
})
class TestAccountProperty(harness.TestCase):
    # TODO(ahuszagh) Check when stabilized.
    pass

    # TODO(ahuszagh)
    #   Need to test a lot of other variants.
    #       Mosaic
    #       Transaction
    #       Ensure the address is correct.


@harness.model_test_case({
    'type': models.AccountProperties,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'address': models.Address('SAUJCIBCOFLHUZIWNB32MR6YUX75HO7GGCVZEXSG'),
        'properties': [
            models.AccountProperty(models.PropertyType.ALLOW_ADDRESS, []),
        ],
    },
    'dto': {
        'accountProperties': {
            'address': '902891202271567A65166877A647D8A5FFD3BBE630AB925E46',
            'properties': [
                {
                    'propertyType': 0x01,
                    'values': [],
                },
            ],
        }
    },
})
class TestAccountProperties(harness.TestCase):
    # TODO(ahuszagh) Check when stabilized.
    pass


@harness.model_test_case({
    'type': models.Address,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'address': 'SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54',
    },
    'extras': {
        'plain': 'SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54',
        'pretty': 'SD5DT3-CH4BLA-BL5HIM-EKP2TA-PUKF4N-Y3L5HR-IR54',
        'encoded': b'\x90\xfa9\xecG\xe0V\x00\xaf\xa7C\x08\xa7\xea`}\x14^7\x1b_O\x14G\xbc',
        'public_key': '7D08373CFFE4154E129E04F0827E5F3D6907587E348757B0F87D2F839BF88246',
        'invalid': 'SD5DT3-CH4BLA-BL5HIM-EKP2TA-PUKF4N-Y3L5HR-IR55'
    }
})
class TestAddress(harness.TestCase):

    def test_properties(self):
        self.assertEqual(self.model.network_type, self.network_type)
        self.assertEqual(self.model.encoded, self.extras['encoded'])

    def test_create_from_raw_address(self):
        self.assertEqual(self.model, self.type.create_from_raw_address(self.extras['pretty']))

    def test_create_from_encoded(self):
        self.assertEqual(self.model, self.type.create_from_encoded(self.extras['encoded']))

    def test_create_from_public_key(self):
        public_key = self.extras['public_key']

        value = self.type.create_from_public_key(public_key, models.NetworkType.MAIN_NET)
        self.assertEqual(value.address, 'XD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L4X2EMBY')

        value = self.type.create_from_public_key(public_key, models.NetworkType.TEST_NET)
        self.assertEqual(value.address, 'VD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L4UGKOZU')

        value = self.type.create_from_public_key(public_key, models.NetworkType.MIJIN)
        self.assertEqual(value.address, 'MD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L4R2R2JH')

        value = self.type.create_from_public_key(public_key, models.NetworkType.MIJIN_TEST)
        self.assertEqual(value.address, 'SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54')

    def test_plain(self):
        self.assertEqual(self.model.plain(), self.extras['plain'])

    def test_pretty(self):
        self.assertEqual(self.model.pretty(), self.extras['pretty'])

    def test_is_valid(self):
        self.assertTrue(self.type.create_from_raw_address(self.extras['plain']).is_valid())
        self.assertTrue(self.type.create_from_raw_address(self.extras['pretty']).is_valid())
        self.assertFalse(self.type.create_from_raw_address(self.extras['invalid']).is_valid())

    def test_invalid_address(self):
        with self.assertRaises(ValueError):
            self.type('')
        with self.assertRaises(ValueError):
            self.type.create_from_encoded('')
        with self.assertRaises(ValueError):
            self.type.create_from_public_key('', self.network_type)


@harness.model_test_case({
    'type': models.MultisigAccountGraphInfo,
    'network_type': models.NetworkType.MIJIN_TEST,
    # Not a dataclass, has no fields or replace methods.
    'fields': False,
    'replace': False,
    'data': {
        0: [
            models.MultisigAccountInfo(
                account=models.PublicAccount(
                    address=models.Address(address='SA2AK4GEVSAX4NNG7KH4EGIXWNHMGEMWU7JT4RU5'),
                    public_key='9b1d7f4f4c2a0471edbc78f95de54b6e432887fc0c6e4ceca800089dae2a4044'
                ),
                min_approval=2,
                min_removal=2,
                cosignatories=[
                    models.PublicAccount(
                        address=models.Address(address='SD5M23VRSNS3AI3NPFUFCIQQTQW2G4L5BLDLMLAJ'),
                        public_key='6cb4caaacca7081c9e1471df8f6512abc99feb86efee7862ed7259397c5fdbdd'
                    ),
                    models.PublicAccount(
                        address=models.Address(address='SBI7LMDC7Y4TDMU3BFNLR7KC7TBACCXBUZ6ZAO6F'),
                        public_key='c5c55181284607954e56cd46de85f4f3ef4cc713cc2b95000fa741998558d268'
                    ),
                    models.PublicAccount(
                        address=models.Address(address='SCSVO5XVNM4AICUEYXMUNEQSW5RUNUOVDLYHBCFM'),
                        public_key='cae725538ebebf7778257a442b4a48e116636580ed630ad5cb8d668dff52a1a7'
                    ),
                ],
                multisig_accounts=[],
            ),
        ],
    },
    'dto': [
        {
            'level': 0,
            'multisigEntries': [
                {
                    'multisig': {
                        'account': '9b1d7f4f4c2a0471edbc78f95de54b6e432887fc0c6e4ceca800089dae2a4044',
                        'accountAddress': '90340570c4ac817e35a6fa8fc21917b34ec31196a7d33e469d',
                        'minApproval': 2,
                        'minRemoval': 2,
                        'cosignatories': [
                            '6cb4caaacca7081c9e1471df8f6512abc99feb86efee7862ed7259397c5fdbdd',
                            'c5c55181284607954e56cd46de85f4f3ef4cc713cc2b95000fa741998558d268',
                            'cae725538ebebf7778257a442b4a48e116636580ed630ad5cb8d668dff52a1a7'
                        ],
                        'multisigAccounts': []
                    },
                },
            ],
        },
    ],
})
class TestMultisigAccountGraphInfo(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.MultisigAccountInfo,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'account': models.PublicAccount.create_from_public_key('9b1d7f4f4c2a0471edbc78f95de54b6e432887fc0c6e4ceca800089dae2a4044', models.NetworkType.MIJIN_TEST),
        'min_approval': 2,
        'min_removal': 2,
        'cosignatories': [
            models.PublicAccount.create_from_public_key('6cb4caaacca7081c9e1471df8f6512abc99feb86efee7862ed7259397c5fdbdd', models.NetworkType.MIJIN_TEST),
            models.PublicAccount.create_from_public_key('c5c55181284607954e56cd46de85f4f3ef4cc713cc2b95000fa741998558d268', models.NetworkType.MIJIN_TEST),
            models.PublicAccount.create_from_public_key('cae725538ebebf7778257a442b4a48e116636580ed630ad5cb8d668dff52a1a7', models.NetworkType.MIJIN_TEST),
        ],
        'multisig_accounts': [],
    },
    'dto': {
        'multisig': {
            'account': '9b1d7f4f4c2a0471edbc78f95de54b6e432887fc0c6e4ceca800089dae2a4044',
            'accountAddress': '90340570c4ac817e35a6fa8fc21917b34ec31196a7d33e469d',
            'minApproval': 2,
            'minRemoval': 2,
            'cosignatories': [
                '6cb4caaacca7081c9e1471df8f6512abc99feb86efee7862ed7259397c5fdbdd',
                'c5c55181284607954e56cd46de85f4f3ef4cc713cc2b95000fa741998558d268',
                'cae725538ebebf7778257a442b4a48e116636580ed630ad5cb8d668dff52a1a7'
            ],
            'multisigAccounts': [],
        },
    },
})
class TestMultisigAccountInfo(harness.TestCase):

    def test_is_multisig(self):
        self.assertTrue(self.model.is_multisig())

    def test_has_cosigner(self):
        for account in self.data['cosignatories']:
            self.assertTrue(self.model.has_cosigner(account))
        for account in self.data['multisig_accounts']:
            self.assertFalse(self.model.has_cosigner(account))
        self.assertFalse(self.model.has_cosigner(self.model.account))

    def test_is_cosigner_of_multisig_account(self):
        for account in self.data['cosignatories']:
            self.assertFalse(self.model.is_cosigner_of_multisig_account(account))
        for account in self.data['multisig_accounts']:
            self.assertTrue(self.model.is_cosigner_of_multisig_account(account))
        self.assertFalse(self.model.is_cosigner_of_multisig_account(self.model.account))


@harness.enum_test_case({
    'type': models.PropertyModificationType,
    'enums': [
        models.PropertyModificationType.ADD,
        models.PropertyModificationType.REMOVE,
    ],
    'values': [
        0x00,
        0x01,
    ],
    'descriptions': [
        "Add property to account",
        "Remove property from account",
    ],
    'dto': [
        0x00,
        0x01,
    ],
    'catbuffer': [
        b'\x00',
        b'\x01',
    ],
})
class TestPropertyModificationType(harness.TestCase):
    pass


@harness.enum_test_case({
    'type': models.PropertyType,
    'enums': [
        models.PropertyType.ALLOW_ADDRESS,
        models.PropertyType.ALLOW_MOSAIC,
        models.PropertyType.ALLOW_TRANSACTION,
        models.PropertyType.SENTINEL,
        models.PropertyType.BLOCK_ADDRESS,
        models.PropertyType.BLOCK_MOSAIC,
        models.PropertyType.BLOCK_TRANSACTION,
    ],
    'values': [
        0x01,
        0x02,
        0x04,
        0x05,
        0x81,
        0x82,
        0x84,
    ],
    'descriptions': [
        "The property type is an address.",
        "The property type is a mosaic id.",
        "The property type is a transaction type.",
        "Property type sentinel.",
        "The property type is a blocking address operation.",
        "The property type is a blocking mosaic id operation.",
        "The property type is a blocking transaction type operation.",
    ],
    'dto': [
        0x01,
        0x02,
        0x04,
        0x05,
        0x81,
        0x82,
        0x84,
    ],
    'catbuffer': [
        b'\x01',
        b'\x02',
        b'\x04',
        b'\x05',
        b'\x81',
        b'\x82',
        b'\x84',
    ],
})
class TestPropertyType(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.PublicAccount,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'public_key': '1b153f8b76ef60a4bfe152f4de3698bd230bac9dc239d4e448715aa46bd58955',
        'address': models.Address('SAUJCIBCOFLHUZIWNB32MR6YUX75HO7GGCVZEXSG'),
    },
    'extras': {
        'message': b'Hello World!',
        'signature': '40af0cb5a7a7533f07a4ba6f1cb2df64f2347feb1b2eaabb9374d28603d146497ea83d3d6ee15758d39c298b48f58e578cc42f36a373e15eef7412e0bd19a801',
        'payload': 'bb000000d0092d8eaf91c07069eeef6651cd313e792b27d2cb31473ceaac40f78ee2121acb5f665826083b87b374c9eb67aefef6b8cf74f0298820a9143b34055e15900c1b153f8b76ef60a4bfe152f4de3698bd230bac9dc239d4e448715aa46bd58955019052420000000000000000f1b4815c00000000009b3155b37159da50aa52d5967c509b410f5a36a3b1e31ecb5ac76675d79b4a5e2000b778a39a3663719dfc5e48c9d78431b1e45c2af9df538782bf199c189dabeac7',
    }
})
class TestPublicAccount(harness.TestCase):

    def test_invalid_public_key(self):
        with self.assertRaises(ValueError):
            self.type(self.data['address'], '')

    def test_properties(self):
        self.assertEqual(self.model.network_type, self.network_type)

    def test_create_from_public_key(self):
        self.assertEqual(self.model, self.type.create_from_public_key(self.data['public_key'], self.network_type))

    @harness.ignore_warnings_test
    def test_verify_signature(self):
        self.assertTrue(self.model.verify_signature(self.extras['message'], self.extras['signature']))
        self.assertFalse(self.model.verify_signature(self.extras['message'], '0' * 128))
        with self.assertRaises(ValueError):
            self.assertTrue(self.model.verify_signature(self.extras['message'], ''))

    @harness.ignore_warnings_test
    def test_verify_transaction(self):
        self.assertTrue(self.model.verify_transaction(self.extras['payload']))



"""
    aggregate_transaction
    =====================

    Transaction containing multiple inner transactions that may be
    initiated by different accounts.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .aggregate_transaction_cosignature import AggregateTransactionCosignature
from .deadline import Deadline
from .inner_transaction import InnerTransaction, InnerTransactionList
from .registry import register_transaction
from .signed_transaction import SignedTransaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.account import Account
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = ['AggregateTransaction', 'AggregateBondedTransaction', 'AggregateCompleteTransaction']


Cosignature = AggregateTransactionCosignature
Cosignatures = typing.Sequence[Cosignature]
TYPES = (
    TransactionType.AGGREGATE_BONDED,
    TransactionType.AGGREGATE_COMPLETE,
)


@util.inherit_doc
@util.dataclass(frozen=True)
class AggregateTransaction(Transaction):
    """
    Transaction containing multiple inner transactions.

    :param network_type: Network type.
    :param type: Transaction type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param inner_transactions: Inner transactions to be included.
    :param signature: (Optional) Transaction signature.
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    inner_transactions: InnerTransactionList

    def __init__(
        self,
        network_type: NetworkType,
        type: TransactionType,
        version: TransactionVersion,
        deadline: Deadline,
        max_fee: int = 0,
        inner_transactions: typing.Optional[InnerTransactionList] = None,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        if type not in TYPES:
            raise ValueError('Invalid transaction type.')
        super().__init__(
            type,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('inner_transactions', inner_transactions or [])

    @classmethod
    def create_complete(
        cls,
        deadline: Deadline,
        inner_transactions: typing.Optional[InnerTransactionList],
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create aggregate complete transaction object.

        :param deadline: Deadline to include transaction.
        :param inner_transactions: Inner transactions to be included.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """

        return cls(
            network_type,
            TransactionType.AGGREGATE_COMPLETE,
            TransactionVersion.AGGREGATE_COMPLETE,
            deadline,
            max_fee,
            inner_transactions,
        )

    @classmethod
    def create_bonded(
        cls,
        deadline: Deadline,
        inner_transactions: typing.Optional[InnerTransactionList],
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create aggregate bonded transaction object.

        :param deadline: Deadline to include transaction.
        :param inner_transactions: Inner transactions to be included.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """

        return cls(
            network_type,
            TransactionType.AGGREGATE_BONDED,
            TransactionVersion.AGGREGATE_BONDED,
            deadline,
            max_fee,
            inner_transactions,
        )

    # SIGNING

#    def sign_with(
#        self,
#        account: Account,
#        gen_hash: typing.AnyStr
#    ) -> SignedTransaction:
#        """
#        Sign transaction.
#
#        :param account: Escrow account.
#        :param gen_hash: Generation hash
#        """
#
#        transaction = self.to_catbuffer()
#        payload = account.sign(transaction, gen_hash)
#        hash = self.transaction_hash(payload, gen_hash)
#
#        return SignedTransaction(
#            payload,
#            hash,
#            account.public_key,
#            self.type,
#            self.network_type
#        )

    def sign_transaction_with_cosignatories(
        self,
        initiator: Account,
        gen_hash: typing.AnyStr,
        cosignatories: typing.Optional[typing.Sequence[Account]] = None,
        fee_strategy: util.FeeCalculationStrategy = util.FeeCalculationStrategy.MEDIUM,
    ) -> SignedTransaction:
        """
        Sign transaction with cosignatories.

        :param initiator_account: Initiator account.
        :param gen_hash: Generation hash
        :param cosignatories: Sequence of accounts cosigning transaction.
        """

        transaction = self.to_catbuffer(fee_strategy=fee_strategy)

        if (cosignatories):
            COSIGNATURE_SIZE = 96
            new_fee = util.calculate_fee(
                fee_strategy,
                self.max_fee,
                self.catbuffer_size() + COSIGNATURE_SIZE * len(cosignatories)
            )

            if (self.max_fee != new_fee):
                transaction = transaction[0:106] + new_fee.to_bytes(8, 'little') + transaction[114:]

        payload = initiator.sign(transaction, gen_hash)  # type: ignore
        hash = self.transaction_hash(payload, gen_hash)  # type: ignore

        if (cosignatories):
            for cosignatory in cosignatories:
                payload += util.decode_hex(cosignatory.public_key)
                payload += cosignatory.sign_data(hash)

            new_size = len(payload)
            payload = new_size.to_bytes(4, 'little') + payload[4:]

        return SignedTransaction(  # type: ignore
            payload,
            hash,
            initiator.public_key,
            self.type,
            self.network_type
        )

    # AGGREGATE

    def to_aggregate(self, signer: PublicAccount):
        raise TypeError('Aggregate transaction cannot be embedded.')

    # CATBUFFER

    def inner_transactions_size(self) -> int:
        """Get payload size, the size in bytes of all sub-transactions."""
        return sum(i.catbuffer_size() for i in self.inner_transactions)

    def catbuffer_size_specific(self) -> int:
        # 4 extra bytes for the payload size.
        # The payload size is the size from all inner transactions.
        extra_size = util.U32_BYTES
        payload_size = self.inner_transactions_size()
        return extra_size + payload_size

    def to_inner_transactions_bytes(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Get the serialized byte array of all sub-transactions."""

        return util.Model.sequence_to_catbuffer(
            self.inner_transactions,
            network_type
        )

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export transfer-specific data to catbuffer."""

        # uint32_t payload_size
        # uint8_t[payload_size] transactions
        payload_size = util.u32_to_catbuffer(self.inner_transactions_size())
        transactions = self.to_inner_transactions_bytes(network_type)
        return payload_size + transactions

    def load_inner_transactions_bytes(
        self,
        data: bytes,
        size: int,
        network_type: NetworkType,
    ) -> bytes:
        """Load inner transactions data from catbuffer."""

        transactions = []
        subdata = data[:size]
        while subdata:
            # This will hard-error if the transaction is invalid,
            # or cut-off, since every deserializer checks the input
            # is valid.
            value, subdata = InnerTransaction.create_from_catbuffer_pair(
                subdata,
                network_type
            )
            transactions.append(value)
        return data[size:]

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load transfer-specific data data from catbuffer."""

        raise NotImplementedError

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'transactions'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        raise NotImplementedError

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        inner_transactions = [InnerTransaction.create_from_dto(x, network_type) for x in data['transactions']]

        self._set('inner_transactions', inner_transactions)


@util.inherit_doc
@register_transaction('AGGREGATE_BONDED')
class AggregateBondedTransaction(AggregateTransaction):

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        inner_transactions: typing.Optional[InnerTransactionList],
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create aggregate bonded transaction object.

        :param deadline: Deadline to include transaction.
        :param inner_transactions: Inner transactions to be included.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls.create_bonded(
            deadline,
            inner_transactions,
            network_type,
            max_fee,
        )


@util.inherit_doc
@register_transaction('AGGREGATE_COMPLETE')
class AggregateCompleteTransaction(AggregateTransaction):

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        inner_transactions: typing.Optional[InnerTransactionList],
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create aggregate complete transaction object.

        :param deadline: Deadline to include transaction.
        :param inner_transactions: Inner transactions to be included.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls.create_complete(
            deadline,
            inner_transactions,
            network_type,
            max_fee,
        )



import binascii

from xpxchain.util.hashlib.keccak import crypto
from xpxchain.util.hashlib.keccak import fallback
from tests import harness
from .helper import *


def generate_testcase(bits, full_hexdigest):
    name = "keccak_{}".format(bits)
    digest_size = bits // 8
    crypto_func = getattr(crypto, name)
    fallback_func = getattr(fallback, name)

    class KeccakTest(harness.TestCase):

        def test_init(self):
            actual = hexdigest(fallback_func(), b'Hello World!')
            expected = hexdigest(crypto_func(), b'Hello World!')
            self.assertEqual(actual, expected)

        def test_properties(self):
            hasher = fallback_func()
            self.assertEqual(hasher.digest_size, digest_size)

        def test_update(self):
            hasher = fallback_func()
            hasher.update(b'Hello')
            hasher.update(b' ')
            hasher.update(b'World!')
            self.assertEqual(hasher.hexdigest(), full_hexdigest)

        def test_digest(self):
            self.assertEqual(digest(fallback_func(), b'Hello World!'), binascii.unhexlify(full_hexdigest))

        def test_hexdigest(self):
            self.assertEqual(hexdigest(fallback_func(), b'Hello World!'), full_hexdigest)

        def test_bytes(self):
            for test in STRING_TESTS:
                actual = hexdigest(fallback_func(), test)
                expected = hexdigest(crypto_func(), test)
                self.assertEqual(actual, expected)

    return KeccakTest


TestKeccak224 = generate_testcase(224, "71519a3ec955d57fce5eabf34f64296e80890478eba9e9b36c9c9d5b")
TestKeccak256 = generate_testcase(256, "3ea2f1d0abf3fc66cf29eebb70cbd4e7fe762ef8a09bcc06c8edf641230afec0")
TestKeccak384 = generate_testcase(384, "1f93aefa2bf7e59893b2f29e0a21a58a7e9bbc3f3ce21f3ab3f7d41aa49fa27ca62fd1f42dc99f8497c346a505154b7e")
TestKeccak512 = generate_testcase(512, "75b70545b09569a8d61251b06fc49b520b6ad5322684fd9466836eb143670afdfa25e0403492e0a7dfb7298a9c7e08576bcf26bc9875adfa88e886009cb2fe00")



# type: ignore
# Keep this import order, since it dramatically speeds up import time.
#   Tested: Python 3.7.2
from .models import *
from .client import *
from .errors import *

__all__ = (
    client.__all__
    + models.__all__
    + errors.__all__
)



"""
    cosignature_signed_transaction
    ==============================

    Cosignature signed transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['CosignatureSignedTransaction']


@util.inherit_doc
@util.dataclass(frozen=True)
class CosignatureSignedTransaction(util.DTO):
    """
    Cosignature signed transaction.

    :param parent_hash: Hash of parent aggregate transaction.
    :param signature: Signature generated by signing the parent hash.
    :param signer: Public key of the transaction signer.
    """

    parent_hash: str
    signature: str
    signer: str

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'parentHash', 'signature', 'signer'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'parentHash': self.parent_hash,
            'signature': self.signature,
            'signer': self.signer,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            parent_hash=data['parentHash'],
            signature=data['signature'],
            signer=data['signer'],
        )



"""
    account
    =======

    Account private key, public key and address.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .address import Address
from .public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util
from ...util.signature import ed25519

__all__ = ['Account']


@util.inherit_doc
@util.dataclass(frozen=True)
class Account(util.Object):
    """
    Describe account via private key, public key and account address.

    :param address: Address for the account.
    :param public_key: Hex-encoded or raw bytes for public key.
    :param private_key: Hex-encoded or raw bytes for private key.
    """

    address: Address
    public_key: str
    private_key: str

    def __init__(
        self,
        address: Address,
        public_key: typing.AnyStr,
        private_key: typing.AnyStr,
    ) -> None:
        public_key = util.encode_hex(public_key)
        private_key = util.encode_hex(private_key)
        if len(public_key) != 64:
            raise ValueError("Invalid public key length")
        if len(private_key) != 64:
            raise ValueError("Invalid private key length")
        self._set('address', address)
        self._set('public_key', public_key)
        self._set('private_key', private_key)

    @property
    def network_type(self) -> NetworkType:
        """Get network type."""
        return self.address.network_type

    @property
    def public_account(self) -> PublicAccount:
        """Get public account."""
        return PublicAccount(self.address, self.public_key)

    @classmethod
    def create_from_private_key(
        cls,
        private_key: typing.AnyStr,
        network_type: NetworkType,
    ):
        """
        Generate Account object from private_key and network type.

        :param private_key: Hex-encoded or raw bytes for private key.
        :param network_type: Network type.
        :return: Account object.
        """

        private_key = util.decode_hex(private_key)
        signing_key = ed25519.sha3.SigningKey(private_key)
        public_key = signing_key.get_verifying_key().to_bytes()
        address = Address.create_from_public_key(public_key, network_type)

        return cls(address, public_key, private_key)  # type: ignore

    @classmethod
    def generate_new_account(
        cls,
        network_type: NetworkType,
        entropy=None
    ):
        """
        Generate new NEM account from network type and random bytes.

        :param network_type: Network type.
        :param entropy: (Optional) callable to generate random bytes for secret key.
        :return: Account object.
        """

        if entropy is not None:
            signing_key, verifying_key = ed25519.sha3.create_keypair(entropy)
        else:
            signing_key, verifying_key = ed25519.sha3.create_keypair()

        public_key = verifying_key.to_bytes()
        private_key = signing_key.to_seed()
        address = Address.create_from_public_key(public_key, network_type)

        return cls(address, public_key, private_key)

    def sign(self, transaction: typing.AnyStr, gen_hash: typing.AnyStr) -> bytes:
        """
        Sign transaction using private key.

        :param private_key: Hex-encoded or raw bytes for transaction data.
        :param gen_hash: Nemesis generation hash.
        :return: Signed transaction data.
        """

        transaction = util.decode_hex(transaction, with_prefix=True)

        # Skip first 100 bytes.
        # uint32_t size
        # uint8_t[64] signature
        # uint8_t[32] signer

        signing_bytes = util.unhexlify(gen_hash) + transaction[100:]

        signature = self.sign_data(signing_bytes)
        public_key = util.unhexlify(self.public_key)
        size = transaction[:4]

        return size + signature + public_key + transaction[100:]

    def sign_data(self, data: typing.AnyStr) -> bytes:
        """
        Sign raw data using private key.

        :param private_key: Hex-encoded or raw bytes to sign.
        :return: Raw data signature.
        """

        data = util.decode_hex(data, with_prefix=True)
        private_key = util.unhexlify(self.private_key)
        public_key = util.unhexlify(self.public_key)
        key = ed25519.sha3.SigningKey(private_key + public_key)

        return typing.cast(bytes, key.sign(data))

    def verify_signature(
        self,
        data: typing.AnyStr,
        signature: typing.AnyStr
    ) -> bool:
        """
        Verify a signature.

        :param data: Hex-encoded or raw bytes used to generate signature.
        :param public_key: Hex-encoded or raw bytes for signature.
        :return: Boolean representing if the signature was verified.
        """
        return self.public_account.verify_signature(data, signature)

    def verify_transaction(
        self,
        transaction: typing.AnyStr
    ) -> bool:
        """
        Verify signed transaction data.

        :param transaction: Hex-encoded or raw bytes for transaction data.
        :return: Boolean representing if the transaction signature was verified.
        """
        return self.public_account.verify_transaction(transaction)



"""
    mosaic_id
    =========

    Identifier for an asset.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from .mosaic_nonce import MosaicNonce
from ..account.public_account import PublicAccount
from ... import util

__all__ = ['MosaicId']


@util.inherit_doc
@util.dataclass(frozen=True, id=0)
class MosaicId(util.IntMixin, util.Object):
    """
    Mosaic identifier.

    :param id: Raw identifier for mosaic.
    """

    id: int

    def __int__(self) -> int:
        return self.id

    def get_id(self) -> str:
        return util.hexlify(self.id.to_bytes(8, 'big'))

    @classmethod
    def create_from_nonce(
        cls,
        nonce: MosaicNonce,
        owner: PublicAccount,
    ):
        """
        Create mosaic ID from nonce and owner.

        :param nonce: Mosaic nonce.
        :param owner: Account of mosaic owner.
        """
        key = util.unhexlify(owner.public_key)
        return cls(util.generate_mosaic_id(nonce.nonce, key))



"""
    modify_metadata_transaction
    ===================================

    Abstract class for modification to account property transactions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .modify_account_metadata_transaction import (
    ModifyAccountMetadataTransaction
)
from .transaction import TransactionBase
from ..metadata.metadata_type import MetadataType
from ..metadata.metadata_modification import MetadataModification
from .recipient import RecipientType
from ..blockchain.network_type import NetworkType

__all__ = ['ModifyMetadataTransaction']

MetadataModificationList = typing.Sequence[MetadataModification]


class ModifyMetadataTransaction(TransactionBase):
    """Abstract class for account property modification transactions."""

    __slots__ = ()

    @staticmethod
    def create_modify_account_metadata_transaction(
        deadline: Deadline,
        metadata_type: MetadataType,
        metadata_id: RecipientType,
        modifications: MetadataModificationList,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new modify account addresses transaction.

        :param deadline: Deadline to include transaction.
        :param property_type: Account property type.
        :param modifications: List of account modifications.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """

        return ModifyAccountMetadataTransaction.create(
            deadline,
            metadata_type,
            metadata_id,
            modifications,
            network_type,
            max_fee,
        )



"""
    format
    ======

    Class-bound helpers to simplify conversion between interchange formats.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .aggregate_transaction_info import AggregateTransactionInfo
from .deadline import Deadline
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util


class FormatBase(util.Object):
    """Utilities to simplify loading and saving to interchange formats."""

    def save(self, key, data, value, network_type) -> None:
        """Save value to interchange format by key."""
        raise util.AbstractMethodError

    def load(self, key, data, network_type):
        """Load value from interchange format by key."""
        raise util.AbstractMethodError

    def load_size(self, data):
        """Load transaction size."""
        return self.load('size', data, None)

    def load_type(self, data):
        """Load transaction type."""
        return self.load('type', data, None)

    def load_network_type(self, data):
        """Load network type."""
        return self.load('network_type', data, None)

    def find_transaction(self, type_map, data):
        """Find derived transaction class via the transaction type."""
        return type_map[self.load_type(data)]


@util.dataclass(frozen=True)
class CatbufferFormat(FormatBase):
    """Utilities to simplify loading and saving to catbuffer."""

    slices: typing.Dict[str, slice]
    size_shared: int

    def save(self, key, data, value, network_type) -> None:
        slc = self.slices[key]
        cb = SAVE_CATBUFFER[key]
        data[slc] = cb(value, network_type)

    def load(self, key, data, network_type):
        slc = self.slices[key]
        cb = LOAD_CATBUFFER[key]
        return cb(data[slc], network_type)


@util.dataclass(frozen=True)
class DTOFormat(FormatBase):
    """Utilities to simplify loading and saving to catbuffer."""

    names: typing.Dict[str, str]

    def save(self, key, data, value, network_type) -> None:
        name = self.names[key]
        cb = SAVE_DTO[key]
        if key != 'transaction_info':
            data = data.setdefault('transaction', {})
        if value is not None:
            data[name] = cb(value, network_type)

    def load(self, key, data, network_type):
        name = self.names[key]
        cb = LOAD_DTO[key]
        if key != 'transaction_info':
            data = data['transaction']
        value = data.get(name)
        if value is not None:
            return cb(value, network_type)


# CATBUFFER HELPERS


def save_signature_catbuffer(signature, network_type):
    if signature is None:
        return bytes(64)
    return util.unhexlify(signature)


def save_signer_catbuffer(signer, network_type):
    if signer is None:
        return bytes(32)
    return util.unhexlify(signer.public_key)


SAVE_CATBUFFER = {
    'size': lambda x, n: util.u32_to_catbuffer(x),
    'signature': save_signature_catbuffer,
    'signer': save_signer_catbuffer,
    'version': lambda x, n: x.to_catbuffer(n),
    'network_type': lambda x, n: x.to_catbuffer(n),
    'type': lambda x, n: x.to_catbuffer(n),
    'max_fee': lambda x, n: util.u64_to_catbuffer(x),
    'deadline': lambda x, n: util.u64_to_catbuffer(x.to_timestamp()),
}


def load_signature_catbuffer(data, network_type):
    if data == bytes(64):
        return None
    return util.hexlify(data)


def load_signer_catbuffer(data, network_type):
    if data == bytes(32):
        return None
    return PublicAccount.create_from_public_key(data, network_type)


LOAD_CATBUFFER = {
    'size': lambda x, n: util.u32_from_catbuffer(x),
    'signature': load_signature_catbuffer,
    'signer': load_signer_catbuffer,
    'version': TransactionVersion.create_from_catbuffer,
    'network_type': NetworkType.create_from_catbuffer,
    'type': TransactionType.create_from_catbuffer,
    'max_fee': lambda x, n: util.u64_from_catbuffer(x),
    'deadline': lambda x, n: Deadline.create_from_timestamp(util.u64_from_catbuffer(x)),
}


# DTO HELPERS


def save_signature_dto(signature, network_type):
    return signature


def save_signer_dto(signer, network_type):
    return signer.public_key


def save_version_dto(version, network_type):
    return version | (int(network_type) << 24)


def save_transaction_info_dto(transaction_info, network_type):
    return transaction_info.to_dto(network_type)


SAVE_DTO = {
    'signature': save_signature_dto,
    'signer': save_signer_dto,
    'version': save_version_dto,
    'type': lambda x, n: x.to_dto(n),
    'max_fee': lambda x, n: util.u64_to_dto(x),
    'deadline': lambda x, n: util.u64_to_dto(x.to_timestamp()),
    'transaction_info': save_transaction_info_dto,
}


def load_signature_dto(data, network_type):
    return data


def load_signer_dto(data, network_type):
    return PublicAccount.create_from_public_key(data, network_type)


def load_version_dto(data, network_type):
    # Version are 3B
    return data & 0xFFFFFF


def load_network_type_dto(data, network_type):
    return NetworkType((data >> 24) & 0x000000FF)


def load_transaction_info_dto(data, network_type):
    if 'hash' in data:
        return TransactionInfo.create_from_dto(data, network_type)
    return AggregateTransactionInfo.create_from_dto(data, network_type)


LOAD_DTO = {
    'signature': load_signature_dto,
    'signer': load_signer_dto,
    'version': load_version_dto,
    'network_type': load_network_type_dto,
    'type': TransactionType.create_from_dto,
    'max_fee': lambda x, n: util.u64_from_dto(x),
    'deadline': lambda x, n: Deadline.create_from_timestamp(util.u64_from_dto(x)),
    'transaction_info': load_transaction_info_dto,
}



def digest(hasher, data):
    hasher.update(data)
    return hasher.digest()


def hexdigest(hasher, data):
    hasher.update(data)
    return hasher.hexdigest()


STRING_TESTS = [
    b'Hello World!',
    # 한국어
    b'\xed\x95\x9c\xea\xb5\xad\xec\x96\xb4',
    # räksmörgås
    b'r\xc3\xa4ksm\xc3\xb6rg\xc3\xa5s',
    # même
    b'm\xc3\xaame',
]



from xpxchain.util import hashlib
from tests import harness


class TestHashFunctions(harness.TestCase):

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_ripemd160(self, data: bytes):
        self.assertEqual(len(hashlib.ripemd160(data).digest()), 20)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_keccak224(self, data: bytes):
        self.assertEqual(len(hashlib.keccak_224(data).digest()), 28)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_keccak256(self, data: bytes):
        self.assertEqual(len(hashlib.keccak_256(data).digest()), 32)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_keccak384(self, data: bytes):
        self.assertEqual(len(hashlib.keccak_384(data).digest()), 48)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_keccak512(self, data: bytes):
        self.assertEqual(len(hashlib.keccak_512(data).digest()), 64)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_sha224(self, data: bytes):
        self.assertEqual(len(hashlib.sha224(data).digest()), 28)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_sha256(self, data: bytes):
        self.assertEqual(len(hashlib.sha256(data).digest()), 32)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_sha384(self, data: bytes):
        self.assertEqual(len(hashlib.sha384(data).digest()), 48)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_sha512(self, data: bytes):
        self.assertEqual(len(hashlib.sha512(data).digest()), 64)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_sha3_224(self, data: bytes):
        self.assertEqual(len(hashlib.sha3_224(data).digest()), 28)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_sha3_256(self, data: bytes):
        self.assertEqual(len(hashlib.sha3_256(data).digest()), 32)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_sha3_384(self, data: bytes):
        self.assertEqual(len(hashlib.sha3_384(data).digest()), 48)

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_sha3_512(self, data: bytes):
        self.assertEqual(len(hashlib.sha3_512(data).digest()), 64)



from xpxchain import models
from tests import harness


@harness.model_test_case({
    'type': models.BlockInfo,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'hash': '04628F15E89936912DCA6345D60E668F513D50E7FC6BAF3F1C723C92CFC4FED6',
        'generation_hash': '7CCDF81A60B0A03A3B30D715C5C7513916319C09215E22C67B0A81106FB21445',
        'total_fee': 0,
        'num_transactions': 31,
        'num_statements': 0,
        'signature': '82D8C518BF32CA9A8478451F79B7EEC69F11E6022AF7230CE21F230A09E51A0CEB169454CB4AD4499BC9C0769B468AB5A6DD716D00C977F16BF63DD1B0653305',
        'network_type': models.NetworkType.MIJIN_TEST,
        'signer': models.PublicAccount(
            public_key='A04335F99D9EE3787528A16C7A302F80D511E9CF71D97D95C2182E0EA75A1EF9',
            address=models.Address('SAHCS4U27RLSCD7N6XWR4KV2M4VEQY2OCXMEZN3H'),
        ),
        'version': models.TransactionVersion.TRANSFER,
        'type': models.BlockType.GENESIS,
        'height': 1,
        'timestamp': 0,
        'difficulty': 100000000000000,
        'fee_multiplier': 0,
        'previous_block_hash': '0000000000000000000000000000000000000000000000000000000000000000',
        'block_transactions_hash': '36822DF1CF1827915A8CDDF8A47EBBF1D1C55666DB1E8AC942C447D3D0E6D494',
        'block_receipts_hash': 'RMViDPTpgKd072qPgxmPVoabWcrmFs5v5EGS00GxiX0=',
        'state_hash': 'C61C28E078298D6A508A649637EAC7D9D1E5F3813856EF9B7F1579866ACB570E',
        'fee_interest': 0,
        'fee_interest_denominator': 0,
    },
    'dto': {
        'meta': {
            'hash': '04628F15E89936912DCA6345D60E668F513D50E7FC6BAF3F1C723C92CFC4FED6',
            'generationHash': '7CCDF81A60B0A03A3B30D715C5C7513916319C09215E22C67B0A81106FB21445',
            'subCacheMerkleRoots': [],
            'totalFee': [0, 0],
            'numTransactions': 31,
            'numStatements': 0,
        },
        'block': {
            'signature': '82D8C518BF32CA9A8478451F79B7EEC69F11E6022AF7230CE21F230A09E51A0CEB169454CB4AD4499BC9C0769B468AB5A6DD716D00C977F16BF63DD1B0653305',
            'signer': 'A04335F99D9EE3787528A16C7A302F80D511E9CF71D97D95C2182E0EA75A1EF9',
            'version': 2415919107,
            'type': 0x8043,
            'height': [1, 0],
            'timestamp': [0, 0],
            'difficulty': [276447232, 23283],
            'feeMultiplier': 0,
            'previousBlockHash': '0000000000000000000000000000000000000000000000000000000000000000',
            'blockTransactionsHash': '36822DF1CF1827915A8CDDF8A47EBBF1D1C55666DB1E8AC942C447D3D0E6D494',
            'blockReceiptsHash': 'RMViDPTpgKd072qPgxmPVoabWcrmFs5v5EGS00GxiX0=',
            'stateHash': 'C61C28E078298D6A508A649637EAC7D9D1E5F3813856EF9B7F1579866ACB570E',
            'feeInterest': 0,
            'feeInterestDenominator': 0
        },
    },
})
class TestBlockInfo(harness.TestCase):

    def test_with_tree(self):
        merkle_tree = [
            "796159C97C02FB6F38BB7C07F9F367A5E7362090D4FDA34739979A6647E127D3",
            "D54C1ED47A4585CA35AB57D8F5FE0C83A7DDBC401B2967C636977E70C3BC5A62",
            "AAFCC40F74636DCCDFE5EABD7FD7BEC5C0C960606E6FFD3BA2515603C4D11272",
            "0000000000000000000000000000000000000000000000000000000000000000",
            "0000000000000000000000000000000000000000000000000000000000000000",
            "0000000000000000000000000000000000000000000000000000000000000000",
            "0000000000000000000000000000000000000000000000000000000000000000"
        ]
        block_info = self.model.replace(merkle_tree=merkle_tree)
        dto = self.dto.copy()
        dto['meta']['subCacheMerkleRoots'] = merkle_tree

        self.assertEqual(dto, block_info.to_dto(self.network_type))
        self.assertEqual(block_info, self.type.create_from_dto(dto, self.network_type))


@harness.enum_test_case({
    'type': models.BlockType,
    'enums': [
        models.BlockType.GENESIS,
        models.BlockType.NEMESIS,
        models.BlockType.REGULAR,
    ],
    'values': [
        0x8043,
        0x8043,
        0x8143,
    ],
    'descriptions': [
        'Genesis',
        'Genesis',
        'Regular',
    ],
    'dto': [
        0x8043,
        0x8043,
        0x8143,
    ],
    'catbuffer': [
        b'C\x80',
        b'C\x80',
        b'C\x81',
    ],
})
class TestBlockType(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.BlockchainScore,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'score': 554597137692201874146690593473,
    },
    'dto': {
        'scoreLow': [2781082305, 27425266],
        'scoreHigh': [5, 7],
    },
})
class TestBlockchainScore(harness.TestCase):

    def test_properties(self):
        self.assertEqual(self.model.score_low, 117790623335183041)
        self.assertEqual(self.model.score_high, 30064771077)


@harness.model_test_case({
    'type': models.BlockchainStorageInfo,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'num_blocks': 11459,
        'num_transactions': 25,
        'num_accounts': 25,
    },
    'dto': {
        'numBlocks': 11459,
        'numTransactions': 25,
        'numAccounts': 25,
    },
})
class TestBlockchainStorageInfo(harness.TestCase):
    pass


@harness.enum_test_case({
    'type': models.NetworkType,
    'enums': [
        models.NetworkType.MAIN_NET,
        models.NetworkType.TEST_NET,
        models.NetworkType.MIJIN,
        models.NetworkType.MIJIN_TEST,
    ],
    'values': [
        0xb8,
        0xa8,
        0x60,
        0x90,
    ],
    'descriptions': [
        "Main network",
        "Test network",
        "Mijin network",
        "Mijin test network",
    ],
    'dto': [
        0xb8,
        0xa8,
        0x60,
        0x90,
    ],
    'catbuffer': [
        b'\xb8',
        b'\xa8',
        b'\x60',
        b'\x90',
    ],
    'custom': [
        {
            'name': 'test_identifier',
            'callback': lambda self, x: x.identifier(),
            'results': [b'X', B'V', b'M', b'S'],
        },
        {
            'name': 'test_create_from_identifier',
            'callback': lambda self, x: self.type.create_from_identifier(x),
            'inputs': [b'X', b'V', b'M', b'S'],
        },
        {
            'name': 'test_create_from_raw_address',
            'callback': lambda self, x: self.type.create_from_raw_address(x),
            'inputs': [
                'XD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54',
                'VD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54',
                'MD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54',
                'SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54',
            ],
        },
        {
            'name': 'test_create_from_invalid_identifier',
            'callback': lambda self, x: self.type.create_from_identifier(x),
            'inputs': [b'F'],
            'results': [KeyError],
        },
        {
            'name': 'test_create_from_invalid_raw_address',
            'callback': lambda self, x: self.type.create_from_raw_address(x),
            'inputs': ['FD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54'],
            'results': [KeyError],
        },
    ],
})
class TestNetworkType(harness.TestCase):
    pass




from __future__ import annotations

from .address_metadata import AddressMetadata
from .mosaic_metadata import MosaicMetadata
from .namespace_metadata import NamespaceMetadata
from .metadata_type import MetadataType
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['MetadataInfo']


@util.inherit_doc
@util.dataclass(frozen=True)
class MetadataInfo(util.DTO):
    """

    :param metadata:  Generic metadata info.

    DTO Format:
        .. code-block:: yaml

            AddressMetadataInfoDTO:
                metedata: MetadataDTO
    """

    metadata: AddressMetadata

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'metadata'}
        required_l2 = {'metadataType', 'fields', 'metadataId'}

        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
            and cls.validate_dto_required(data['metadata'], required_l2)
            and cls.validate_dto_all(data['metadata'], required_l2)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'metadata': self.metadata.to_dto(),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        metadata = data['metadata']

        if (metadata['metadataType'] == MetadataType.ADDRESS):
            return cls(
                metadata=AddressMetadata.create_from_dto(metadata)
            )
        elif (metadata['metadataType'] == MetadataType.MOSAIC):
            return cls(
                metadata=MosaicMetadata.create_from_dto(metadata)
            )
        elif (metadata['metadataType'] == MetadataType.NAMESPACE):
            return cls(
                metadata=NamespaceMetadata.create_from_dto(metadata)
            )
        else:
            raise ValueError('Invalid data-transfer object.')



import binascii

from xpxchain import util
from tests import harness


class Base32Test(harness.TestCase):

    def test_b32encode(self):
        self.assertEqual(util.b32encode(b'5'), 'GU======')
        self.assertEqual(util.b32encode(b'5', with_suffix=False), 'GU')

        with self.assertRaises(TypeError):
            util.b32encode('5')

    def test_encode_base32(self):
        self.assertEqual(util.encode_base32(b'5'), 'GU======')
        self.assertEqual(util.encode_base32('GU======'), 'GU======')
        self.assertEqual(util.encode_base32('GU======', with_suffix=False), 'GU')

    def test_b32decode(self):
        self.assertEqual(util.b32decode('GU======'), b'5')
        self.assertEqual(util.b32decode(b'GU======'), b'5')
        self.assertEqual(util.b32decode('GU', with_suffix=False), b'5')
        self.assertEqual(util.b32decode(b'GU', with_suffix=False), b'5')

        with self.assertRaises(binascii.Error):
            util.b32decode('G', with_suffix=False)
        with self.assertRaises(binascii.Error):
            util.b32decode('GU')

    def test_decode_base32(self):
        self.assertEqual(util.decode_base32('GU======'), b'5')
        self.assertEqual(util.decode_base32(b'5'), b'5')


class Base64Test(harness.TestCase):

    def test_b64encode(self):
        self.assertEqual(util.b64encode(b'5'), 'NQ==')
        self.assertEqual(util.b64encode(b'5', with_suffix=False), 'NQ')

        with self.assertRaises(TypeError):
            util.b64encode('5')

    def test_encode_base64(self):
        self.assertEqual(util.encode_base64(b'5'), 'NQ==')
        self.assertEqual(util.encode_base64('NQ=='), 'NQ==')
        self.assertEqual(util.encode_base64('NQ==', with_suffix=False), 'NQ')

    def test_b64decode(self):
        self.assertEqual(util.b64decode('NQ=='), b'5')
        self.assertEqual(util.b64decode(b'NQ=='), b'5')
        self.assertEqual(util.b64decode('NQ', with_suffix=False), b'5')
        self.assertEqual(util.b64decode(b'NQ', with_suffix=False), b'5')

        with self.assertRaises(binascii.Error):
            util.b64decode('N', with_suffix=False)
        with self.assertRaises(binascii.Error):
            util.b64decode('NQ')

    def test_decode_base64(self):
        self.assertEqual(util.decode_base64('NQ=='), b'5')
        self.assertEqual(util.decode_base64(b'5'), b'5')



"""
    signed_transaction
    ==================

    Signed transaction data.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .transaction_type import TransactionType
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = ['SignedTransaction']

AnyStr1 = typing.TypeVar('AnyStr1', bytes, str)
AnyStr2 = typing.TypeVar('AnyStr2', bytes, str)
AnyStr3 = typing.TypeVar('AnyStr3', bytes, str)


@util.inherit_doc
@util.dataclass(frozen=True)
class SignedTransaction(util.Object):
    """
    Signed transaction data and signature/hash.

    :param payload: Serialized transaction data.
    :param hash: Transaction hash.
    :param signer: Transaction signer (public key).
    :param type: Transaction type.
    :param network_type: Signer network type.
    """

    payload: str
    hash: str
    signer: str
    type: TransactionType
    network_type: NetworkType

    def __init__(
        self,
        payload: typing.AnyStr,
        hash: typing.AnyStr,
        signer: typing.AnyStr,
        type: TransactionType,
        network_type: NetworkType,
    ) -> None:
        payload = util.encode_hex(payload)
        hash = util.encode_hex(hash)
        signer = util.encode_hex(signer)
        if len(hash) != 64:
            raise ValueError('Transaction hash must be 64 characters long.')
        if len(signer) not in (0, 64):
            raise ValueError('Signer public key must be empty or 64 characters long.')
        self._set('payload', payload)
        self._set('hash', hash)
        self._set('signer', signer)
        self._set('type', type)
        self._set('network_type', network_type)

    @classmethod
    def create_from_announced(
        cls,
        hash: typing.AnyStr,
        type: TransactionType,
        network_type: NetworkType,
    ):
        """
        Create minimal signed transaction from announced transaction hash and type.

        :param hash: Transaction hash.
        :param type: Transaction type.
        :param network_type: Signer network type.
        """
        return cls('', hash, '', type, network_type)  # type: ignore

    def __eq__(self, other) -> bool:
        # Since we may use `create_from_announced` and we want it
        # to compare equal, we ignore payload and signer in equality checks.
        if not isinstance(other, SignedTransaction):
            return False
        lhs = (self.hash, self.type, self.network_type)
        rhs = (other.hash, other.type, other.network_type)
        return lhs == rhs



import binascii

from xpxchain import util
from tests import harness


class TestBinascii(harness.TestCase):

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_hex_encoding(self, decoded: bytes):
        encoded = util.hexlify(decoded)
        self.assertEqual(len(encoded), 2 * len(decoded))
        self.assertEqual(util.encode_hex(decoded), encoded)
        self.assertEqual(util.encode_hex(encoded), encoded)
        self.assertEqual(util.unhexlify(encoded), decoded)
        self.assertEqual(util.decode_hex(decoded), decoded)
        self.assertEqual(util.decode_hex(encoded), decoded)

        with self.assertRaises(binascii.Error):
            util.decode_hex(encoded + '0')
        with self.assertRaises(binascii.Error):
            util.decode_hex(encoded[:-1])



"""
    recipient
    =========

    Recipient utilities to handle both addresses and namespace IDs.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from ..account.address import Address
from ..blockchain.network_type import NetworkType
from ..namespace.namespace_id import NamespaceId
from ... import util
from ...util import bit

RecipientType = typing.Union[Address, NamespaceId]


class Recipient(util.Object):
    """Wrapper for the recipient DTO and catbuffer objects."""

    CATBUFFER_SIZE: typing.ClassVar[int] = 25

    @staticmethod
    def to_dto(
        obj: RecipientType,
        network_type: NetworkType,
    ) -> str:
        return util.hexlify(Recipient.to_catbuffer(obj, network_type))

    @staticmethod
    def create_from_dto(
        data: str,
        network_type: NetworkType,
    ) -> RecipientType:
        return Recipient.create_from_catbuffer(util.unhexlify(data), network_type)

    @staticmethod
    def to_catbuffer(
        obj: RecipientType,
        network_type: NetworkType,
    ) -> bytes:
        if isinstance(obj, Address):
            # The first byte is always the network type.
            return obj.encoded
        # The first byte is always the network type + 1.
        leading = util.u8_to_catbuffer(int(network_type) + 1)
        trailing = bytes(16)
        return leading + util.unhexlify(obj.encoded) + trailing

    @staticmethod
    def create_from_catbuffer(
        data: bytes,
        network_type: NetworkType,
    ) -> RecipientType:
        # For addresses, the first byte is the network type,
        # which are always even. If it's odd, we have a namespace ID.
        # Otherwise, for the namespace ID, the first byte is a sentinel,
        # and the next 8 bytes are the relevant data.
        # The first byte without the first bit is always the network type.
        assert bit.clear(data[0], 0) == int(network_type)
        if bit.get(data[0], 0) == 0:
            return typing.cast(Address, Address.create_from_encoded(data[:25]))
        return typing.cast(NamespaceId, NamespaceId.create_from_encoded(data[1:9]))

    @staticmethod
    def create_from_catbuffer_pair(
        data: bytes,
        network_type: NetworkType,
    ) -> typing.Tuple[RecipientType, bytes]:
        size = Recipient.CATBUFFER_SIZE
        inst = Recipient.create_from_catbuffer(data, network_type)
        return inst, data[size:]



"""
    mosaic_definition_transaction
    =============================

    Mosaic-definition transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ..mosaic.mosaic_id import MosaicId
from ..mosaic.mosaic_nonce import MosaicNonce
from ..mosaic.mosaic_properties import MosaicProperties, MosaicDefinitionProperties
from ... import util

__all__ = [
    'MosaicDefinitionTransaction',
    'MosaicDefinitionInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('MOSAIC_DEFINITION')
class MosaicDefinitionTransaction(Transaction):
    """
    Mosaic-definition transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param nonce: Mosaic nonce (random data for mosaic uniqueness).
    :param mosaic_id: Identifier for mosaic.
    :param mosaic_properties: Mosaic properties.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    nonce: MosaicNonce
    mosaic_id: MosaicId
    mosaic_properties: MosaicProperties

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        nonce: MosaicNonce,
        mosaic_id: MosaicId,
        mosaic_properties: MosaicProperties,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        super().__init__(
            TransactionType.MOSAIC_DEFINITION,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('nonce', nonce)
        self._set('mosaic_id', mosaic_id)
        self._set('mosaic_properties', mosaic_properties)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        nonce: MosaicNonce,
        mosaic_id: MosaicId,
        mosaic_properties: MosaicProperties,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new mosaic definition transaction.

        :param deadline: Deadline to include transaction.
        :param nonce: Mosaic nonce (random data for mosaic uniqueness).
        :param mosaic_id: Identifier for mosaic.
        :param mosaic_properties: Mosaic properties.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.MOSAIC_DEFINITION,
            deadline,
            nonce,
            mosaic_id,
            mosaic_properties,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        # TODO(ahuszagh) Likely have to remove the Nonce, because
        # it likely does not have an actual format.
        nonce_size = MosaicNonce.CATBUFFER_SIZE
        id_size = util.U64_BYTES
        definition_properties = MosaicDefinitionProperties(self.mosaic_properties)
        properties_size = definition_properties.catbuffer_size()
        return nonce_size + id_size + properties_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export mosaic definition-specific data to catbuffer."""

        # MosaicNonce nonce
        # uint64_t mosaic_id
        # MosaicProperties properties
        nonce = self.nonce.to_catbuffer(network_type)
        mosaic_id = util.u64_to_catbuffer(int(self.mosaic_id))
        definition_properties = MosaicDefinitionProperties(self.mosaic_properties)
        properties = definition_properties.to_catbuffer(network_type)
        return nonce + mosaic_id + properties

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load mosaic definition-specific data data from catbuffer."""

        # MosaicNonce nonce
        # MosaicId mosaic_id
        # MosaicProperties properties
        nonce, data = MosaicNonce.create_from_catbuffer_pair(data, network_type)
        mosaic_id = MosaicId(util.u64_from_catbuffer(data[:8]))
        properties, data = MosaicDefinitionProperties.create_from_catbuffer_pair(
            data[8:],
            network_type
        )

        self._set('nonce', nonce)
        self._set('mosaic_id', mosaic_id)
        self._set('mosaic_properties', properties.model)

        return typing.cast(bytes, data)

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'mosaicNonce', 'mosaicId', 'properties'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        definition_properties = MosaicDefinitionProperties(self.mosaic_properties)
        return {
            'mosaicNonce': self.nonce.to_dto(network_type),
            'mosaicId': util.u64_to_dto(int(self.mosaic_id)),
            'properties': definition_properties.to_dto(network_type),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        nonce = MosaicNonce.create_from_dto(data['mosaicNonce'], network_type)
        mosaic_id = MosaicId(util.u64_from_dto(data['mosaicId']))
        properties = MosaicDefinitionProperties.create_from_dto(
            data['properties'],
            network_type
        )
        self._set('nonce', nonce)
        self._set('mosaic_id', mosaic_id)
        self._set('mosaic_properties', properties.model)


@register_transaction('MOSAIC_DEFINITION')
class MosaicDefinitionInnerTransaction(InnerTransaction, MosaicDefinitionTransaction):
    """Embedded mosaic definition transaction."""

    __slots__ = ()



"""
    modify_account_property_transaction
    ===================================

    Abstract class for modification to account property transactions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .modify_account_property_address_transaction import (
    ModifyAccountPropertyAddressTransaction
)
from .modify_account_property_entity_type_transaction import (
    ModifyAccountPropertyEntityTypeTransaction
)
from .modify_account_property_mosaic_transaction import (
    ModifyAccountPropertyMosaicTransaction
)
from .account_property_modification import AccountPropertyModification
from .transaction import TransactionBase
from .transaction_type import TransactionType
from ..account.address import Address
from ..account.property_modification_type import PropertyModificationType
from ..account.property_type import PropertyType
from ..blockchain.network_type import NetworkType
from ..mosaic.mosaic_id import MosaicId

__all__ = ['AccountPropertyTransaction']

AccountPropertyModificationList = typing.Sequence[AccountPropertyModification]


class AccountPropertyTransaction(TransactionBase):
    """Abstract class for account property modification transactions."""

    __slots__ = ()

    @staticmethod
    def create_address_property_modification_transaction(
        deadline: Deadline,
        property_type: PropertyType,
        modifications: AccountPropertyModificationList,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new modify account addresses transaction.

        :param deadline: Deadline to include transaction.
        :param property_type: Account property type.
        :param modifications: List of account modifications.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """

        return ModifyAccountPropertyAddressTransaction.create(
            deadline,
            property_type,
            modifications,
            network_type,
            max_fee,
        )

    @staticmethod
    def create_mosaic_property_modification_transaction(
        deadline: Deadline,
        property_type: PropertyType,
        modifications: AccountPropertyModificationList,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new modify account mosaics transaction.

        :param deadline: Deadline to include transaction.
        :param property_type: Account property type.
        :param modifications: List of account modifications.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """

        return ModifyAccountPropertyMosaicTransaction.create(
            deadline,
            property_type,
            modifications,
            network_type,
            max_fee,
        )

    @staticmethod
    def create_entity_type_property_modification_transaction(
        deadline: Deadline,
        property_type: PropertyType,
        modifications: AccountPropertyModificationList,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new modify account entity types transaction.

        :param deadline: Deadline to include transaction.
        :param property_type: Account property type.
        :param modifications: List of account modifications.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """

        return ModifyAccountPropertyEntityTypeTransaction.create(
            deadline,
            property_type,
            modifications,
            network_type,
            max_fee,
        )

    @staticmethod
    def create_address_filter(
        modification_type: PropertyModificationType,
        address: Address,
    ) -> AccountPropertyModification:
        """
        Create address filter for account property modification.

        :param modification_type: Add or remove modification.
        :param address: Address value for modification.
        """
        return AccountPropertyModification(modification_type, address)

    @staticmethod
    def create_mosaic_filter(
        modification_type: PropertyModificationType,
        mosaic_id: MosaicId,
    ) -> AccountPropertyModification:
        """
        Create mosaic filter for account property modification.

        :param modification_type: Add or remove modification.
        :param mosaic_id: Mosaic ID value for modification.
        """
        return AccountPropertyModification(modification_type, mosaic_id)

    @staticmethod
    def create_entity_type_filter(
        modification_type: PropertyModificationType,
        entity_type: TransactionType,
    ) -> AccountPropertyModification:
        """
        Create entity type filter for account property modification.

        :param modification_type: Add or remove modification.
        :param entity_type: Transaction type value for modification.
        """
        return AccountPropertyModification(modification_type, entity_type)



import binascii

from Crypto.Hash import SHA3_224, SHA3_256, SHA3_384, SHA3_512
from xpxchain.util import hashlib
from tests import harness
from .helper import *

HASH = {
    224: SHA3_224,
    256: SHA3_256,
    384: SHA3_384,
    512: SHA3_512,
}


def generate_testcase(bits, full_hexdigest):
    name = "sha3_{}".format(bits)
    digest_size = bits // 8
    func = getattr(hashlib, name)
    func_c = HASH[bits].new

    class Sha3Test(harness.TestCase):

        def test_init(self):
            actual = hexdigest(func(), b'Hello World!')
            expected = hexdigest(func_c(), b'Hello World!')
            self.assertEqual(actual, expected)

        def test_properties(self):
            hasher = func()
            self.assertEqual(hasher.digest_size, digest_size)

        def test_update(self):
            hasher = func()
            hasher.update(b'Hello')
            hasher.update(b' ')
            hasher.update(b'World!')
            self.assertEqual(hasher.hexdigest(), full_hexdigest)

        def test_digest(self):
            self.assertEqual(digest(func(), b'Hello World!'), binascii.unhexlify(full_hexdigest))

        def test_hexdigest(self):
            self.assertEqual(hexdigest(func(), b'Hello World!'), full_hexdigest)

        def test_bytes(self):
            for test in STRING_TESTS:
                actual = hexdigest(func(), test)
                expected = hexdigest(func_c(), test)
                self.assertEqual(actual, expected)

    return Sha3Test


TestSha3224 = generate_testcase(224, "716596afadfa17cd1cb35133829a02b03e4eed398ce029ce78a2161d")
TestSha3256 = generate_testcase(256, "d0e47486bbf4c16acac26f8b653592973c1362909f90262877089f9c8a4536af")
TestSha3384 = generate_testcase(384, "f324cbd421326a2abaedf6f395d1a51e189d4a71c755f531289e519f079b224664961e385afcc37da348bd859f34fd1c")
TestSha3512 = generate_testcase(512, "32400b5e89822de254e8d5d94252c52bdcb27a3562ca593e980364d9848b8041b98eabe16c1a6797484941d2376864a1b0e248b0f7af8b1555a778c336a5bf48")



"""
    register_namespace_transaction
    ==============================

    Register namespace transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ..namespace.namespace_id import NamespaceId
from ..namespace.namespace_type import NamespaceType
from ... import util

__all__ = [
    'RegisterNamespaceTransaction',
    'RegisterNamespaceInnerTransaction',
]


def to_namespace_id(namespace: typing.Union[str, NamespaceId]) -> NamespaceId:
    """Convert namespace name or ID to namespace ID."""

    if isinstance(namespace, str):
        return NamespaceId(namespace)
    return namespace


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('REGISTER_NAMESPACE')
class RegisterNamespaceTransaction(Transaction):
    """
    Register namespace transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param namespace_type: Root or sub namespace.
    :param namespace_name: Name of namespace to register.
    :param namespace_id: ID of namespace to register.
    :param duration: (Optional) Duration of the namespace.
    :param parent_id: (Optional) Parent namespace ID.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    namespace_type: NamespaceType
    namespace_name: str
    namespace_id: NamespaceId
    duration: typing.Optional[int]
    parent_id: typing.Optional[NamespaceId]

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        namespace_type: NamespaceType,
        namespace_name: str,
        namespace_id: NamespaceId,
        max_fee: int = 0,
        duration: typing.Optional[int] = None,
        parent_id: typing.Optional[NamespaceId] = None,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        if namespace_type == NamespaceType.ROOT_NAMESPACE and duration is None:
            raise ValueError('Registering a root namespace without a duration.')
        elif namespace_type == NamespaceType.SUB_NAMESPACE and parent_id is None:
            raise ValueError('Registering a sub namespace without a parent.')
        if duration is not None and parent_id is not None:
            raise ValueError('Cannot have both namespace duration and parent.')
        super().__init__(
            TransactionType.REGISTER_NAMESPACE,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('namespace_type', namespace_type)
        self._set('namespace_name', namespace_name)
        self._set('namespace_id', namespace_id)
        self._set('duration', duration)
        self._set('parent_id', parent_id)

    @classmethod
    def create_root_namespace(
        cls,
        deadline: Deadline,
        namespace_name: str,
        duration: int,
        network_type: NetworkType,
    ):
        """
        Create new root namespace transaction.

        :param deadline: Deadline to include transaction.
        :param namespace_name: Name of the namespace to register.
        :param duration: Duration of the namespace.
        :param network_type: Network type.
        """
        return cls(
            network_type,
            TransactionVersion.REGISTER_NAMESPACE,
            deadline,
            NamespaceType.ROOT_NAMESPACE,
            namespace_name,
            NamespaceId(namespace_name),
            duration=duration,
            parent_id=None,
        )

    @classmethod
    def create_sub_namespace(
        cls,
        deadline: Deadline,
        namespace_name: str,
        parent_namespace: typing.Union[str, NamespaceId],
        network_type: NetworkType,
    ):
        """
        Create new sub namespace transaction.

        Multiple parent names may be provided using a `.` delimiter to
        separate parent namespaces, up to a depth of 2, or a parent
        namespace identifier may be provided instead.

        :param deadline: Deadline to include transaction.
        :param namespace_name: Name of the namespace to register.
        :param parent_namespace: Name or identifier of the parent namespace.
        :param network_type: Network type.
        """
        parent_id = to_namespace_id(parent_namespace)
        id = util.generate_sub_namespace_id(parent_id.id, namespace_name)
        return cls(
            network_type,
            TransactionVersion.REGISTER_NAMESPACE,
            deadline,
            NamespaceType.SUB_NAMESPACE,
            namespace_name.split('.')[-1],
            NamespaceId(id),
            duration=None,
            parent_id=parent_id
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        # Duration size includes both either the parent ID or the duration.
        # We have 1 extra byte, as a marker for the size of the namespace name.
        extra_size = util.U8_BYTES
        namespace_type_size = util.U8_BYTES
        duration_or_id_size = util.U64_BYTES
        namespace_id_size = util.U64_BYTES
        namespace_name_size = util.U8_BYTES * len(self.namespace_name)
        return (
            extra_size
            + namespace_type_size
            + duration_or_id_size
            + namespace_id_size
            + namespace_name_size
        )

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export register namespace-specific data to catbuffer."""

        # uint8_t namespace_type
        # uint64_t duration || parent_id
        # uint64_t namespace_id
        # uint8_t namespace_name_size
        # uint8_t[namespace_name_size] namespace_name
        namespace_type = self.namespace_type.to_catbuffer(network_type)
        if self.duration is not None:
            duration_or_id = util.u64_to_catbuffer(self.duration)
        else:
            duration_or_id = util.u64_to_catbuffer(int(self.parent_id))
        namespace_id = util.u64_to_catbuffer(int(self.namespace_id))
        namespace_name_size = util.u8_to_catbuffer(len(self.namespace_name))
        namespace_name = self.namespace_name.encode('ascii')
        namespace = namespace_id + namespace_name_size + namespace_name

        return namespace_type + duration_or_id + namespace

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load transfer-specific data from catbuffer."""

        # uint8_t namespace_type
        # uint64_t duration || parent_id
        # uint64_t namespace_id
        # uint8_t namespace_name_size
        # uint8_t[namespace_name_size] namespace_name
        namespace_type, data = NamespaceType.create_from_catbuffer_pair(
            data,
            network_type
        )
        if namespace_type == NamespaceType.ROOT_NAMESPACE:
            duration = util.u64_from_catbuffer(data[:8])
            parent_id = None
        else:
            duration = None
            parent_id = NamespaceId(util.u64_from_catbuffer(data[:8]))
        namespace_id = NamespaceId(util.u64_from_catbuffer(data[8:16]))
        namespace_name_size = util.u8_from_catbuffer(data[16:17])
        data = data[17:]
        namespace_name = data[:namespace_name_size].decode('ascii')
        data = data[namespace_name_size:]

        self._set('namespace_type', namespace_type)
        self._set('namespace_name', namespace_name)
        self._set('namespace_id', namespace_id)
        self._set('duration', duration)
        self._set('parent_id', parent_id)

        return typing.cast(bytes, data)

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'namespaceType', 'name', 'namespaceId'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        data = {
            'namespaceType': self.namespace_type.to_dto(network_type),
            'name': self.namespace_name,
            'namespaceId': util.u64_to_dto(int(self.namespace_id)),
        }

        if self.duration is not None:
            data['duration'] = util.u64_to_dto(self.duration)
        if self.parent_id is not None:
            data['parentId'] = util.u64_to_dto(int(self.parent_id))

        return data

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        namespace_type = NamespaceType.create_from_dto(
            data['namespaceType'],
            network_type
        )
        namespace_name = data['name']
        namespace_id = NamespaceId(util.u64_from_dto(data['namespaceId']))
        if namespace_type == NamespaceType.ROOT_NAMESPACE:
            duration = util.u64_from_dto(data['duration'])
            parent_id = None
        else:
            duration = None
            parent_id = NamespaceId(util.u64_from_dto(data['parentId']))

        self._set('namespace_type', namespace_type)
        self._set('namespace_name', namespace_name)
        self._set('namespace_id', namespace_id)
        self._set('duration', duration)
        self._set('parent_id', parent_id)


@register_transaction('REGISTER_NAMESPACE')
class RegisterNamespaceInnerTransaction(InnerTransaction, RegisterNamespaceTransaction):
    """Embedded transfer transaction."""

    __slots__ = ()



"""
    documentation
    =============

    Decorators to modify doc strings.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import inspect
import typing

__all__ = [
    'inherit_doc',
]


def is_public(f):
    """Check if function is public."""
    return not f.__name__.startswith('_')


def has_doc(f):
    """Check if function has doc string."""
    return f.__doc__ is not None


def needs_doc(f):
    """Check if function is public and has no doc string."""
    return is_public(f) and not has_doc(f)


def inherit_doc(cls: typing.Type) -> typing.Type:
    """Inherit documentation from a base class."""

    # Ignore anything coming from Python internally, or that
    # is private to begin with.
    bases = [i for i in cls.__bases__ if i is not object]
    members = inspect.getmembers(cls, inspect.isroutine)
    members = ((k, v) for k, v in members if needs_doc(v))
    for key, func in members:
        for base in bases:
            doc = getattr(getattr(base, key, None), '__doc__', None)
            if doc is not None:
                # Methods use a property to get the doc.
                if inspect.ismethod(func):
                    func.__func__.__doc__ = doc
                else:
                    func.__doc__ = doc

    return cls



"""
    transaction_type
    ================

    Enumerations for transaction types.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import enum

from ... import util

__all__ = ['TransactionType']


@util.inherit_doc
class TransactionType(util.U16Mixin, util.EnumMixin, enum.IntEnum):
    """Transaction type."""

    TRANSFER = 0x4154
    REGISTER_NAMESPACE = 0x414E
    MOSAIC_DEFINITION = 0x414D
    MOSAIC_SUPPLY_CHANGE = 0x424D
    MODIFY_MULTISIG_ACCOUNT = 0x4155
    AGGREGATE_COMPLETE = 0x4141
    AGGREGATE_BONDED = 0x4241
    LOCK = 0x4148
    SECRET_LOCK = 0x4152
    SECRET_PROOF = 0x4252
    ADDRESS_ALIAS = 0x424E
    MOSAIC_ALIAS = 0x434E
    MODIFY_ACCOUNT_PROPERTY_ADDRESS = 0x4150
    MODIFY_ACCOUNT_PROPERTY_MOSAIC = 0x4250
    MODIFY_ACCOUNT_PROPERTY_ENTITY_TYPE = 0x4350
    LINK_ACCOUNT = 0x414C
    MODIFY_ACCOUNT_METADATA = 0x413D
    MODIFY_MOSAIC_METADATA = 0x423D
    MODIFY_NAMESPACE_METADATA = 0x433D
    BLOCKCHAIN_UPGRADE = 0x4158
    NETWORK_CONFIG = 0x4159

    def description(self) -> str:
        return DESCRIPTION[self]


DESCRIPTION = {
    TransactionType.TRANSFER: "Transfer transaction type.",
    TransactionType.REGISTER_NAMESPACE: "Register namespace transaction type.",
    TransactionType.MOSAIC_DEFINITION: "Mosaic definition transaction type.",
    TransactionType.MOSAIC_SUPPLY_CHANGE: "Mosaic supply change transaction.",
    TransactionType.MODIFY_MULTISIG_ACCOUNT: "Modify multisig account transaction type.",
    TransactionType.AGGREGATE_COMPLETE: "Aggregate complete transaction type.",
    TransactionType.AGGREGATE_BONDED: "Aggregate bonded transaction type.",
    TransactionType.LOCK: "Lock transaction type.",
    TransactionType.SECRET_LOCK: "Secret lock transaction type.",
    TransactionType.SECRET_PROOF: "Secret proof transaction type.",
    TransactionType.ADDRESS_ALIAS: "Address alias transaction type.",
    TransactionType.MOSAIC_ALIAS: "Mosaic alias transaction type.",
    TransactionType.MODIFY_ACCOUNT_PROPERTY_ADDRESS: "Modify account property address transaction type.",  # noqa: E501
    TransactionType.MODIFY_ACCOUNT_PROPERTY_MOSAIC: "Modify account property mosaic transaction type.",    # noqa: E501
    TransactionType.MODIFY_ACCOUNT_PROPERTY_ENTITY_TYPE: "Modify account property entity type transaction type.",  # noqa: E501
    TransactionType.LINK_ACCOUNT: "Link account transaction type.",
    TransactionType.MODIFY_ACCOUNT_METADATA: "Modify account metadata transaction type",
    TransactionType.MODIFY_MOSAIC_METADATA: "Modify mosaic metadata transaction type",
    TransactionType.MODIFY_NAMESPACE_METADATA: "Modify namespace metadata transaction type",
    TransactionType.BLOCKCHAIN_UPGRADE: "Blockchain upgrade transaction",
    TransactionType.NETWORK_CONFIG: "Network configuration transaction",
}
